program main

  use vtype_module

  use fi_module

  use constants0, only : PI

  use rad_planetconst, only : &
    & RadPlanetConstInit, &
    & Grav,            &
!!$    & Cp,              &
    & MolWtNonRadAct

  use gauss_quad, only : GauLeg

  use rad_flux, only : RadFluxCalcFluxMultiSZA, RadFluxInit, RadFluxFinalize, RadFluxOutputInit, RadFluxOutputFinalize
  use rad_kd_utils, only : RadKDUtilsCalcTendencyVarCp
  use rad_output, only : RadOutputFluxTendency

  use hitranconst, only : m_HITRANMolWt

  use convection, only : &
    & ConvCalcRadTendency, &
    & ConvCalcRadConvTendency

  use io_profile, only : &
    & IOProfileInqDimLen, &
    & IOProfileRead, &
    & IOProfileReadPtcl, &
    & IOProfileOutput

  use cloud_majcomp, only : CloudMajComp

  use saturation_K1991, only : SatK1991CalcTempSat

#ifdef THERMOPROP_VIRACpLSF
  use thermo_prop_co2_viracplsf, only : &
#else
  use thermo_prop_co2_sw1996, only : &
#endif
    & ThermoPropInit


  implicit none


  character(stdstr) :: CtlFN0 = 'planetconst.conf'
  real(DP)          :: GravSet
  real(DP)          :: CpSet
  real(DP)          :: MolWtNonRadActSet

  character(extstr) :: CtlFN1 = 'thermo_prop.conf'
  character(extstr) :: CtlFN3 = 'calc_rte.conf'
  character(extstr) :: CtlFN4 = 'calc_rce.conf'


  character(len=extstr*4) :: GasOptPropNcFN
  character(len=extstr*4) :: PtclOptPropNcFN
  character(len=extstr*4) :: KDOptPropNcFN

  character(len=extstr*4)   :: InitFileName
  character(len=extstr*4)   :: InFileName
  character(len=extstr)     :: PressName
  character(len=extstr)     :: MolNumName
  character(len=extstr)     :: TempName
  character(len=extstr)     :: SurfTempName
  character(len=extstr)     :: VMRName


  logical           :: FlagOutSpeFlux = .false.
  character(stdstr) :: OutSpeFluxNcFN
  character(stdstr) :: OutRadFluxNcFN
  character(stdstr) :: OutRadTendencyNcFN
  character(stdstr) :: OutTendenciesNcFN

  real(dp)       :: FilterCoef

  real(DP), allocatable :: z_Cp       (:)

  integer           :: iostat

  real(dp), allocatable :: r_Sigma(:)
  real(dp), allocatable :: z_DelAtmMass(:)
  real(dp), allocatable :: rm_MMR        (:,:)
  real(dp), allocatable :: r_MMRNonRadAct(:)
  real(dp), allocatable :: zm_MMR        (:,:)
  real(dp), allocatable :: z_MMRNonRadAct(:)
  real(dp), allocatable :: m_AtmMassRadAct(:)
  real(dp)              :: AtmMassDry
  real(dp)              :: AtmMassNonRadAct
  real(dp)              :: AtmMassTot
  real(dp)              :: AtmMassDrySave
  real(dp)              :: AtmMassNonRadActSave
  real(dp), allocatable :: m_AtmMassRadActSave(:)

  real(dp), allocatable :: r_DelTempRC    (:)
  real(dp)              :: DelSurfTemp
  real(dp), allocatable :: r_DelTempFilter    (:)
  real(dp), allocatable :: r_DelVMRH2OVap     (:)
  real(dp), allocatable :: r_Height           (:)


  integer, parameter :: MaxNmlArySize = 64


  integer                       :: kmax

  !
  ! NMolNum : number of radiatively active species
  !
  integer                        :: NMolNum


  real(dp)                           :: DelWaveNum
  real(dp)                           :: WaveNumS
  real(dp)                           :: WaveNumE


  integer(i4b)              :: FileUnit

  logical                   :: FlagConvection
  logical                   :: FlagMoistConvection
  logical                   :: FlagMajCompCond
  logical                   :: FlagMajCompCloudEvolve
  logical                   :: FlagMajCompCloudSedimentation
  logical                   :: FlagMajCompCloudSedimentationInfty
  character(stdstr)         :: MajCompCloudName
  real(dp)                  :: MajCompCloudDens


  real(dp)    , allocatable :: r_Press (:)
  integer     , allocatable :: m_MolNum(:)
  real(dp)    , allocatable :: rm_VMR  (:,:)

  real(dp)    , allocatable :: r_Temp(:)
  real(dp)                  :: SurfTemp
  real(dp)    , allocatable :: z_Press(:)

  real(dp)    , allocatable :: r_Enthalpy     (:)
  real(dp)    , allocatable :: r_DelEnthalpy  (:)
  real(dp)    , allocatable :: r_TempTentative(:)

  real(dp)    , allocatable :: r_TempB(:)
  real(dp)                  :: SurfTempB

  integer              , parameter   :: NChar = stdstr
  integer                            :: NPtcl
  integer              , allocatable :: a_PtclNum     (:)
  character(len=NChar) , allocatable :: a_PtclName    (:)
  real(8)              , allocatable :: za_PtclDens   (:,:)
  real(8)              , allocatable :: za_PtclEffRad (:,:)
  real(8)              , allocatable :: za_PtclMMR    (:,:)
  real(8)              , allocatable :: za_DelPtclMass(:,:)
  real(8)              , allocatable :: za_DelPtclMassTentative(:,:)
  real(8)              , allocatable :: za_PtclMMRforRad(:,:)

  real(8)              , allocatable :: a_PtclDens    (:)

  real(8), allocatable :: r_MeanMolWt(:)
  real(8), allocatable :: r_VMRNonRadAct(:)
  real(8), allocatable :: z_DTempDt      (:)
  real(8), allocatable :: z_DTempDtLW    (:)
  real(8), allocatable :: z_DTempDtSW    (:)

  integer               :: NDelFlxMax
  real(DP), allocatable :: r_RadUwFlux    (:)
  real(DP), allocatable :: r_RadDwFlux    (:)
  real(DP), allocatable :: ra_DelRadUwFlux(:,:)
  real(DP), allocatable :: ra_DelRadDwFlux(:,:)

  real(DP), allocatable :: r_RadUwFluxLW    (:)
  real(DP), allocatable :: r_RadDwFluxLW    (:)
  real(DP), allocatable :: r_RadUwFluxSW    (:)
  real(DP), allocatable :: r_RadDwFluxSW    (:)
  real(DP), allocatable :: ra_DelRadUwFluxLW(:,:)
  real(DP), allocatable :: ra_DelRadDwFluxLW(:,:)
  real(DP), allocatable :: ra_DelRadUwFluxSW(:,:)
  real(DP), allocatable :: ra_DelRadDwFluxSW(:,:)

  real(DP), allocatable :: r_RadUwFluxEach      (:)
  real(DP), allocatable :: r_RadDwFluxEach      (:)
  real(DP), allocatable :: ra_DelRadUwFluxEach  (:,:)
  real(DP), allocatable :: ra_DelRadDwFluxEach  (:,:)
  real(DP), allocatable :: r_RadUwFluxLWEach    (:)
  real(DP), allocatable :: r_RadDwFluxLWEach    (:)
  real(DP), allocatable :: r_RadUwFluxSWEach    (:)
  real(DP), allocatable :: r_RadDwFluxSWEach    (:)
  real(DP), allocatable :: ra_DelRadUwFluxLWEach(:,:)
  real(DP), allocatable :: ra_DelRadDwFluxLWEach(:,:)
  real(DP), allocatable :: ra_DelRadUwFluxSWEach(:,:)
  real(DP), allocatable :: ra_DelRadDwFluxSWEach(:,:)

  real(DP), allocatable :: rs_RadUwFluxEach      (:,:)
  real(DP), allocatable :: rs_RadDwFluxEach      (:,:)
  real(DP), allocatable :: ras_DelRadUwFluxEach  (:,:,:)
  real(DP), allocatable :: ras_DelRadDwFluxEach  (:,:,:)
  real(DP), allocatable :: rs_RadUwFluxSWEach    (:,:)
  real(DP), allocatable :: rs_RadDwFluxSWEach    (:,:)
  real(DP), allocatable :: ras_DelRadUwFluxSWEach(:,:,:)
  real(DP), allocatable :: ras_DelRadDwFluxSWEach(:,:,:)

  real(8), allocatable :: r_RadUwFluxTentative  (:)
  real(8), allocatable :: r_RadDwFluxTentative  (:)
  real(8), allocatable :: r_RadUwFluxLWTentative(:)
  real(8), allocatable :: r_RadDwFluxLWTentative(:)
  real(8), allocatable :: r_RadUwFluxSWTentative(:)
  real(8), allocatable :: r_RadDwFluxSWTentative(:)

  real(DP)              :: StrFluxTOA
  character(len=extstr) :: StrSpeNcFN

  real(DP) :: AlbedoSW
  real(DP) :: AlbedoLW
  real(DP) :: EmisSW
  real(DP) :: EmisLW
  real(DP) :: WaveNumAlbedoSwitch

  integer              , parameter   :: NCloudPtclMax = 256
  character(len=extstr)              :: CloudPtclNames(NCloudPtclMax)
  real(DP)                           :: CloudFraction
  integer                            :: NCloudPtcl
  character(len=extstr), allocatable :: a_CloudPtclNames(:)

  integer               :: NGQ
  real(DP), allocatable :: a_GQMu    (:)
  real(DP), allocatable :: a_GQWeight(:)

  integer               :: NGQLoop
!!$  real(DP)              :: SZADegEach
  real(DP), allocatable :: s_SZADeg(:)

  real(DP) :: SZADeg

  real(DP) :: TempSat

  real(DP) :: WNIntegStart
  real(DP) :: WNIntegEnd


  character(extstr)       :: OutFN
  character(extstr)       :: OutDir
  character(extstr)       :: OutNcFN = 'profile.nc'

  real(DP), allocatable   :: r_Dens     (:)
  real(DP), allocatable   :: r_AdLapRate(:)

!!$  logical  :: FlagUpdateSurfTemp = .true.
  logical  :: FlagAdjustAtmMass  = .false.
  real(DP) :: SensHeatFlux
!!$  real(DP) :: SurfNetUwFlux
  real(DP) :: SurfHeatCap

!!$  real(DP), allocatable :: r_DelAtmMass(:)
  real(DP)              :: DelMajCompCldMass
  real(DP)              :: CloudNumRatio
  real(DP)              :: MajCompLatentHeat

  integer :: IndexMajCompCloud

  integer  :: NTime
  real(DP) :: DelTime

  integer  :: iCldClr
  integer  :: iCldClrS
  integer  :: iCldClrE
  real(dp) :: FluxFactor

  integer :: k
  integer :: m
  integer :: n
  integer :: iTime
  integer :: OutputStep
  integer :: OutputStdOutStep
  logical :: FlagStdOut

  logical  :: FlagIdealGasSet
  real(dp) :: ConvThresholdSet


  namelist /planetconst_nml/ &
    & GravSet, CpSet, MolWtNonRadActSet
  namelist /wn_nml/ &
    & DelWaveNum, WaveNumS, WaveNumE

  namelist /rce_nml/ &
    & KDOptPropNcFN, &
    & FlagOutSpeFlux, &
    & DelTime, &
    & SurfHeatCap, &
    & NTime, &
    & OutputStep, OutputStdOutStep, &
    & FlagConvection, &
    & FlagMoistConvection, &
    & FlagMajCompCond, &
    & FlagMajCompCloudEvolve, &
    & FlagMajCompCloudSedimentation, &
    & FlagMajCompCloudSedimentationInfty, &
    & MajCompCloudName, &
    & MajCompCloudDens, &
    & MajCompLatentHeat, &
    & CloudNumRatio, &
    & NCloudPtcl, CloudPtclNames, CloudFraction, &
    & FilterCoef, &
    & InitFileName, &
    & PressName, MolNumName, TempName, VMRName, &
    & OutDir
!!$    & SurfTempName

  namelist /rte_nml/ &
    & StrFluxTOA, &
    & StrSpeNcFN, &
    & AlbedoSW, &
    & AlbedoLW, &
    & WaveNumAlbedoSwitch, &
    & NGQ, &
    & SZADeg, &
    & WNIntegStart, &
    & WNIntegEnd

  namelist /thermo_prop_nml/ &
    & FlagIdealGasSet, ConvThresholdSet


  NTime       = 100
  DelTime     = 84600.0d0 * 10
  OutputStep       = 1
  OutputStdOutStep = 1000
  SurfHeatCap = 4.217d3 * 1.0d3 * 10.0d0  ! [J K-1 kg-1] * [kg m-3] * [m]

  KDOptPropNcFN      = ''
  OutRadFluxNcFN     = 'out/flux.nc'
  OutRadTendencyNcFN = 'out/tendency.nc'
  FlagOutSpeFlux     = .false.
  OutSpeFluxNcFN     = 'out/flux_spectrum.nc'
  OutTendenciesNcFN  = 'out/tendencies.nc'

  FlagConvection                = .true.
  FlagMoistConvection           = .true.

  FlagMajCompCond                    = .false.
  FlagMajCompCloudEvolve             = .true.
  FlagMajCompCloudSedimentation      = .true.
  FlagMajCompCloudSedimentationInfty = .true.
  MajCompCloudName       = ''
  MajCompCloudDens       = 1.62d3 ! CO2 density at 150 K (Mangan et al., 2017)
  MajCompLatentHeat      = 5.98d5 ! Stull (1947)
  CloudNumRatio          = 1.0d7

  NCloudPtcl     = 0
  CloudPtclNames = ""
  CloudFraction  = 1.0_DP

  FilterCoef             = 0.0d0

  !======================
  InitFileName = "./data.nc.nc"
  PressName    = "Press"
  MolNumName   = "MolNum"
  TempName     = "Temp"
  SurfTempName = "Ts"
  VMRName      = "VMR"
  !======================

  OutDir = "out"


!!$  StrFluxTOA          = 1370.0d0
  StrFluxTOA          =    0.0d0
  StrSpeNcFN          = "strspe.nc"
  AlbedoSW            = 0.2d0
  AlbedoLW            = 0.0d0
  WaveNumAlbedoSwitch = 3250.0d2
  NGQ                 =   -1
  SZADeg              =    0.0d0

  WNIntegStart        = -1.0d0
  WNIntegEnd          = -1.0d0

  FlagIdealGasSet  = .false.
  ConvThresholdSet = 1.0d-11


  ! Read a namelist file

  call fi_open( CtlFN0, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = planetconst_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist, planetconst_nml, is something wrong.'
  end if
  write( 6, nml = planetconst_nml )
  !
  close( FileUnit )

  call fi_open( CtlFN1, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = thermo_prop_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in thermo_prop_nml is something wrong.'
  end if
  write( 6, nml = thermo_prop_nml )
  !
  close( FileUnit )

  call fi_open( CtlFN4, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = rce_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in rce_nml is something wrong.'
  end if
  write( 6, nml = rce_nml )
  !
  close( FileUnit )

  call fi_open( CtlFN3, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = rte_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in rte_nml is something wrong.'
  end if
  write( 6, nml = rte_nml )
  !
  close( FileUnit )


  ! Initialization of thermodynamic properties
#ifdef THERMOPROP_VIRACpLSF
  call ThermoPropInit(     &
    & ConvThresholdSet = ConvThresholdSet &
!!$    ! If specific heat is constant, a following line can be used.
!!$    ! Do not forget to modify argument of RadPlanetConstInit in that case.
!!$    & ConvThresholdSet = ConvThresholdSet, ConstCp = CpSet &
    & )
#else
  call ThermoPropInit( &
    & FlagIdealGasSet, ConvThresholdSet  &
    & )
#endif


  EmisSW = 1.0d0 - AlbedoSW
  EmisLW = 1.0d0 - AlbedoLW

  if ( ( CloudFraction < 0.0_DP ) .or. ( 1.0_DP < CloudFraction ) ) then
    write( 6, * ) 'CloudFraction is out of range: ', CloudFraction
    stop
  end if
  if ( NCloudPtcl > NCloudPtclMax ) then
    write( 6, * ) 'NCloudPtcl is too large: ', NCloudPtcl, NCloudPtclMax
    stop
  end if
  allocate( a_CloudPtclNames(NCloudPtcl) )
  a_CloudPtclNames(1:NCloudPtcl) = CloudPtclNames(1:NCloudPtcl)


  ! Initialize planetary constants
  !
  call RadPlanetConstInit( &
    & GravSet           = GravSet,           &
    ! In default setting, a constant value of Cp is not used in this program.
    ! If specific heat is constant, a following line can be used.
    ! Do not forget to modify argument of ThermoPropInit in that case.
!!$    & CpSet             = CpSet,             &
    & MolWtNonRadActSet = MolWtNonRadActSet  &
    & )


  InFileName = InitFileName



  call IOProfileInqDimLen(      &
    & InFileName, PressName, &
    & kmax &
    & )
  kmax = kmax - 1
  call IOProfileInqDimLen(      &
    & InFileName, MolNumName, &
    & NMolNum &
    & )
  allocate( r_Press (0:kmax   ) )
  allocate( m_MolNum(1:NMolNum) )
  allocate( r_Temp  (0:kmax   ) )
  allocate( rm_VMR  (0:kmax   ,1:NMolNum) )
  call IOProfileRead(      &
    & InFileName, PressName, MolNumName, TempName, SurfTempName, VMRName, &
    & kmax+1, NMolNum, &
    & r_Press, m_MolNum, r_Temp, SurfTemp, rm_VMR     &
    & )

!!$  SurfTemp = r_Temp(0)


  allocate( r_MeanMolWt   (0:kmax) )
  allocate( r_VMRNonRadAct(0:kmax) )

  allocate( r_DelTempRC    (0:kmax) )
  allocate( r_DelTempFilter(0:kmax) )
  allocate( r_DelVMRH2OVap (0:kmax) )

  allocate( r_Sigma     (0:kmax) )
  allocate( z_DelAtmMass(1:kmax) )
  allocate( rm_MMR        (0:kmax,1:NMolNum) )
  allocate( r_MMRNonRadAct(0:kmax) )
  allocate( zm_MMR        (1:kmax,1:NMolNum) )
  allocate( z_MMRNonRadAct(1:kmax) )
  allocate( m_AtmMassRadAct    (1:NMolNum) )
  allocate( m_AtmMassRadActSave(1:NMolNum) )
  !
  r_Sigma = r_Press / r_Press(0)


  do k = 1, kmax
    z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
  end do
!!$  AtmMassTot = 0.0d0
!!$  do k = 1, kmax
!!$    AtmMassTot = AtmMassTot + z_DelAtmMass(k)
!!$  end do

  ! volume mixing ratio of non-radiative active specie
  r_VMRNonRadAct = 1.0d0
  do m = 1, NMolNum
    r_VMRNonRadAct = r_VMRNonRadAct - rm_VMR(:,m)
  end do
  ! mean molecular weight
  call CalcMeanMolWt( &
    & kmax, NMolNum, m_MolNum, rm_VMR, r_VMRNonRadAct, &
    & r_MeanMolWt, &
!!$    & .true. & ! FlagOutput
    & .false. &
    & )
  ! mass mixing ratio
  !   non-radiatively active specie
  r_MMRNonRadAct = r_VMRNonRadAct * MolWtNonRadAct / r_MeanMolWt
  do k = 1, kmax
    z_MMRNonRadAct(k) = ( r_MMRNonRadAct(k-1) + r_MMRNonRadAct(k) ) / 2.0d0
  end do
  !   radiatively active specie
  do m = 1, NMolNum
    rm_MMR(:,m) = rm_VMR(:,m) * m_HITRANMolWt(m_MolNum(m)) / r_MeanMolWt
  end do
  do m = 1, NMolNum
    do k = 1, kmax
      zm_MMR(k,m) = ( rm_MMR(k-1,m) + rm_MMR(k,m) ) / 2.0d0
    end do
  end do
  ! column mass
  !   non-radiatively active specie
  AtmMassNonRadAct = 0.0d0
  do k = 1, kmax
    AtmMassNonRadAct = AtmMassNonRadAct &
      & + z_DelAtmMass(k) * z_MMRNonRadAct(k)
  end do
  !   radiatively active specie
  m_AtmMassRadAct = 0.0d0
  do m = 1, NMolNum
    do k = 1, kmax
      m_AtmMassRadAct(m) = m_AtmMassRadAct(m) &
        & + z_DelAtmMass(k) * zm_MMR(k,m)
    end do
  end do
  !   total
  AtmMassTot = AtmMassNonRadAct
  do m = 1, NMolNum
    AtmMassTot = AtmMassTot + m_AtmMassRadAct(m)
  end do
  !   dry
  AtmMassDry = AtmMassTot - m_AtmMassRadAct(1)
  ! save
  AtmMassNonRadActSave = AtmMassNonRadAct
  m_AtmMassRadActSave  = m_AtmMassRadAct
  AtmMassDrySave       = AtmMassDry


  ! temporarily set
!!$  NPtcl              = 2
!!$  do m = 1, NPtcl
!!$    a_PtclNum(m) = m
!!$  end do
  !
  call IOProfileInqDimLen(      &
    & InFileName, "PtclNum", &
    & NPtcl &
    & )
  !
  allocate( z_Press     (1:kmax) )
  allocate( r_Height    (0:kmax) )
  !
  allocate( a_PtclNum (NPtcl) )
  allocate( a_PtclName(NPtcl) )
  allocate( za_PtclEffRad (kmax,NPtcl) )
  allocate( za_PtclDens   (kmax,NPtcl) )
  allocate( za_PtclMMR    (kmax,NPtcl) )
  allocate( za_DelPtclMass(kmax,NPtcl) )
  allocate( za_DelPtclMassTentative(kmax,NPtcl) )
  allocate( za_PtclMMRforRad(kmax,NPtcl) )
  !
  allocate( a_PtclDens    (NPtcl) )
  !
!!$  do k = 1, kmax
!!$    if ( r_Press(k-1) == 0.0d0 ) then
!!$      z_Press(k) = ( r_Press(k-1) + r_Press(k) ) / 2.0d0
!!$    else
!!$      z_Press(k) = sqrt( r_Press(k-1) * r_Press(k) )
!!$    end if
!!$  end do
  !
!!$  a_PtclName(1)      = "water"
!!$  a_PtclName(2)      = "ice"
!!$  za_PtclEffRad(:,1) = 10.0d-6
!!$  za_PtclEffRad(:,2) = 50.0d-6
!!$  za_DelPtclNum      = 0.0d0
  !
  call IOProfileReadPtcl(      &
    & InFileName, &
    & kmax, NPtcl, &
    & z_Press, a_PtclNum, a_PtclName, a_PtclDens, za_PtclEffRad, za_PtclMMR  &
    & )
  do m = 1, NPtcl
    za_PtclDens(:,m) = a_PtclDens(m)
  end do


  allocate( r_Enthalpy     (0:kmax) )
  allocate( r_DelEnthalpy  (0:kmax) )
  allocate( r_TempTentative(0:kmax) )


  allocate( r_TempB      (0:kmax) )

  allocate( z_DTempDt  (1:kmax) )
  allocate( z_DTempDtLW(1:kmax) )
  allocate( z_DTempDtSW(1:kmax) )

  NDelFlxMax = 1

  allocate( r_RadUwFlux(0:kmax) )
  allocate( r_RadDwFlux(0:kmax) )
  allocate( ra_DelRadUwFlux(0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadDwFlux(0:kmax,-1:NDelFlxMax) )

  allocate( r_RadUwFluxLW    (0:kmax) )
  allocate( r_RadDwFluxLW    (0:kmax) )
  allocate( r_RadUwFluxSW    (0:kmax) )
  allocate( r_RadDwFluxSW    (0:kmax) )
  allocate( ra_DelRadUwFluxLW(0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadDwFluxLW(0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadUwFluxSW(0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadDwFluxSW(0:kmax,-1:NDelFlxMax) )

  if ( NGQ > 0 ) then
    NGQLoop = NGQ + 1
    allocate( a_GQMu    ( NGQ+1 ) )
    allocate( a_GQWeight( NGQ+1 ) )
    call GauLeg( 0.0d0, 1.0d0, NGQ, a_GQMu(1:NGQ), a_GQWeight(1:NGQ) )
    ! divid weights by 2 in order to consider night side
    a_GQWeight = a_GQWeight / 2.0d0
    ! night side
    a_GQMu    (NGQ+1) = -1.0d0
    a_GQWeight(NGQ+1) =  0.5d0
  else
    NGQLoop = 1
    allocate( a_GQMu    ( 1 ) )
    allocate( a_GQWeight( 1 ) )
    n = 1
    a_GQMu    (n) = cos( SZADeg * PI / 180.0d0 )
    a_GQWeight(n) = 1.0d0
  end if

  allocate( r_RadUwFluxEach      (0:kmax) )
  allocate( r_RadDwFluxEach      (0:kmax) )
  allocate( ra_DelRadUwFluxEach  (0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadDwFluxEach  (0:kmax,-1:NDelFlxMax) )
  allocate( r_RadUwFluxLWEach    (0:kmax) )
  allocate( r_RadDwFluxLWEach    (0:kmax) )
  allocate( r_RadUwFluxSWEach    (0:kmax) )
  allocate( r_RadDwFluxSWEach    (0:kmax) )
  allocate( ra_DelRadUwFluxLWEach(0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadDwFluxLWEach(0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadUwFluxSWEach(0:kmax,-1:NDelFlxMax) )
  allocate( ra_DelRadDwFluxSWEach(0:kmax,-1:NDelFlxMax) )
  !
  allocate( s_SZADeg              (1:NGQLoop) )
  allocate( rs_RadUwFluxEach      (0:kmax,1:NGQLoop) )
  allocate( rs_RadDwFluxEach      (0:kmax,1:NGQLoop) )
  allocate( ras_DelRadUwFluxEach  (0:kmax,-1:NDelFlxMax,1:NGQLoop) )
  allocate( ras_DelRadDwFluxEach  (0:kmax,-1:NDelFlxMax,1:NGQLoop) )
  allocate( rs_RadUwFluxSWEach    (0:kmax,1:NGQLoop) )
  allocate( rs_RadDwFluxSWEach    (0:kmax,1:NGQLoop) )
  allocate( ras_DelRadUwFluxSWEach(0:kmax,-1:NDelFlxMax,1:NGQLoop) )
  allocate( ras_DelRadDwFluxSWEach(0:kmax,-1:NDelFlxMax,1:NGQLoop) )
  !
  allocate( r_RadUwFluxTentative  (0:kmax) )
  allocate( r_RadDwFluxTentative  (0:kmax) )
  allocate( r_RadUwFluxLWTentative(0:kmax) )
  allocate( r_RadDwFluxLWTentative(0:kmax) )
  allocate( r_RadUwFluxSWTentative(0:kmax) )
  allocate( r_RadDwFluxSWTentative(0:kmax) )
  !
!!$  allocate( r_DelAtmMass(0:kmax) )


  allocate( z_Cp       (1:kmax) )

  allocate( r_Dens     (0:kmax) )
  allocate( r_AdLapRate(0:kmax) )


  iTime = 0
  write( OutFN, '(a,i6.6,a)' ) &
    & trim( OutDir ) // "/" // OutNcFN(1:len_trim(OutNcFN)-3), iTime, '.nc'
!!$  call Output( &
!!$    & OutFN, &
!!$    & kmax, r_Press, r_Temp, NMolNum, m_MolNum, rm_VMR &
!!$    & )
  do m = 1, NPtcl
    a_PtclDens(m) = za_PtclDens(1,m)
  end do
  call CalcAdditionalVars(   &
    & Grav, kmax, r_Press, r_Temp, & ! (in)
    & r_Dens, r_AdLapRate          & ! (out)
    & )
  call IOProfileOutput( &
    & OutFN, &
    & kmax+1, NMolNum, r_Press, m_MolNum, &
    & r_Temp, rm_VMR, r_Press(0), SurfTemp, &
    & z_Press, &
    & NChar, NPtcl, a_PtclNum, a_PtclName, &
    & a_PtclDens, za_PtclEffRad, za_PtclMMR, &
    & r_Dens, r_AdLapRate                    &
    & )

  PtclOptPropNcFN = ''


  if ( FlagMajCompCond ) then
    do m = 1, NPtcl
      if ( trim( a_PtclName(m) ) == trim( MajCompCloudName ) ) exit
    end do
    if ( m > NPtcl ) then
      write( 6, * ) 'Unable to find ', trim( MajCompCloudName ), ' in PtclName.'
      write( 6, * ) '   ', a_PtclName
      stop
    end if
    IndexMajCompCloud = m
  else
    IndexMajCompCloud = -1
  end if

  do iTime = 1, NTime
    if ( mod( iTime, OutputStdOutStep ) == 1 ) then
      FlagStdOut = .true.
    else
      FlagStdOut = .false.
    end if

    if ( FlagStdOut ) then
      write( 6, * ) 'Loop: ', iTime
      call flush( 6 )
    end if

    r_TempB   = r_Temp
    SurfTempB = SurfTemp

    call CalcMeanMolWt( &
      & kmax, NMolNum, m_MolNum, rm_VMR, r_VMRNonRadAct, &
      & r_MeanMolWt, &
      & .false. &
      & )

    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do
    AtmMassTot = 0.0d0
    do k = 1, kmax
      AtmMassTot = AtmMassTot + z_DelAtmMass(k)
    end do
    !
    r_MMRNonRadAct = r_VMRNonRadAct * MolWtNonRadAct / r_MeanMolWt
    do m = 1, NMolNum
      rm_MMR(:,m) = rm_VMR(:,m) * m_HITRANMolWt(m_MolNum(m)) / r_MeanMolWt
    end do
    do k = 1, kmax
      z_MMRNonRadAct(k) = ( r_MMRNonRadAct(k-1) + r_MMRNonRadAct(k) ) / 2.0d0
    end do
    do m = 1, NMolNum
      do k = 1, kmax
        zm_MMR(k,m) = ( rm_MMR(k-1,m) + rm_MMR(k,m) ) / 2.0d0
      end do
    end do
    AtmMassNonRadAct = 0.0d0
    do k = 1, kmax
      AtmMassNonRadAct = AtmMassNonRadAct &
        & + z_DelAtmMass(k) * z_MMRNonRadAct(k)
    end do
    m_AtmMassRadAct = 0.0d0
    do m = 1, NMolNum
      do k = 1, kmax
        m_AtmMassRadAct(m) = m_AtmMassRadAct(m) &
          & + z_DelAtmMass(k) * zm_MMR(k,m)
      end do
    end do


    GasOptPropNcFN  = KDOptPropNcFN


    za_PtclMMRforRad = za_PtclMMR
    do m = 1, NPtcl
      do k = 1, kmax
!!$        za_DelPtclMass(k,m) = za_PtclMMR(k,m) &
!!$          & * ( r_Press(k-1) - r_Press(k) ) / Grav
        za_DelPtclMass(k,m) = za_PtclMMRforRad(k,m) &
          & * ( r_Press(k-1) - r_Press(k) ) / Grav
      end do
    end do


    if ( iTime == 1 ) then
      call RadFluxInit( &
        & Grav, &
        & WaveNumAlbedoSwitch, &
        & GasOptPropNcFN, PtclOptPropNcFN, &
        & StrSpeNcFN, &
        & Method = "KD" &
        & )
    end if

    if ( FlagOutSpeFlux ) then
      call RadFluxOutputInit( &
        & GasOptPropNcFN, &
        & OutSpeFluxNcFN, &
        & kmax, r_Press &
        & )
    end if
    if ( iTime == 1 ) then
      if ( NGQ > 0 ) then
        write( 6, '(i3,a)' ) NGQ, '-point Gaussian quadrature for dayside + nightside:'
        call flush( 6 )
      end if
    end if
    !
    s_SZADeg = acos( a_GQMu ) * 180.0d0/PI
    !
    r_RadUwFlux       = 0.0_DP
    r_RadDwFlux       = 0.0_DP
    ra_DelRadUwFlux   = 0.0_DP
    ra_DelRadDwFlux   = 0.0_DP
    r_RadUwFluxLW     = 0.0_DP
    r_RadDwFluxLW     = 0.0_DP
    r_RadUwFluxSW     = 0.0_DP
    r_RadDwFluxSW     = 0.0_DP
    ra_DelRadUwFluxLW = 0.0_DP
    ra_DelRadDwFluxLW = 0.0_DP
    ra_DelRadUwFluxSW = 0.0_DP
    ra_DelRadDwFluxSW = 0.0_DP
    !
    ! iCldClr = 1 : cloudy sky part
    ! iCldClr = 2 : clear sky part
    if ( 0.0_DP < CloudFraction ) then
      iCldClrS = 1  ! cloudy sky part will be calculated
    else
      iCldClrS = 2  ! cloudy sky part will NOT be calculated
    end if
    if ( CloudFraction < 1.0_DP ) then
      iCldClrE = 2  ! clear sky part will be calculated
    else
      iCldClrE = 1  ! clear sky part will NOT be calculated
    end if
    do iCldClr = iCldClrS, iCldClrE
      r_RadUwFluxEach       = 0.0d0
      r_RadDwFluxEach       = 0.0d0
      ra_DelRadUwFluxEach   = 0.0d0
      ra_DelRadDwFluxEach   = 0.0d0
      r_RadUwFluxSWEach     = 0.0d0
      r_RadDwFluxSWEach     = 0.0d0
      ra_DelRadUwFluxSWEach = 0.0d0
      ra_DelRadDwFluxSWEach = 0.0d0
      select case ( iCldClr )
      case ( 1 )
        ! set mass mixing ratios for cloud particle for flux calculation
        za_DelPtclMassTentative = za_DelPtclMass
      case ( 2 )
        ! set zero mass mixing ratio for cloud particle
        do m = 1, NPtcl
          do n = 1, NCloudPtcl
            if ( a_PtclName(m) == a_CloudPtclNames(n) ) then
              za_DelPtclMassTentative(:,m) = 0.0_DP
            end if
          end do
        end do
      case default
        write( 6, * ) 'ERROR : Unexpected number of iCldClr, iCldClr'
        stop
      end select
      call RadFluxCalcFluxMultiSZA( &
        & kmax, NMolNum, &
        & m_MolNum, &
        & r_Press, &
        & SurfTemp, r_Temp, rm_VMR, &
!        & NPtcl, a_PtclName, za_PtclDens, za_PtclEffRad, za_DelPtclMass, &
        & NPtcl, a_PtclName, za_PtclDens, za_PtclEffRad, za_DelPtclMassTentative, &
        & AlbedoSW, AlbedoLW, &
        & EmisSW, EmisLW, &
        & NGQLoop, s_SZADeg, &
        & StrFluxTOA, &
        & r_MeanMolWt, &
        & NDelFlxMax, &
        & rs_RadUwFluxEach, rs_RadDwFluxEach, &
        & ras_DelRadUwFluxEach, ras_DelRadDwFluxEach, &
        & r_RadUwFluxLWEach , r_RadDwFluxLWEach,  &
        & rs_RadUwFluxSWEach, rs_RadDwFluxSWEach, &
        & ra_DelRadUwFluxLWEach , ra_DelRadDwFluxLWEach,  &
        & ras_DelRadUwFluxSWEach, ras_DelRadDwFluxSWEach, &
        & WNIntegStart, WNIntegEnd &
        & )
      do n = 1, NGQLoop
        r_RadUwFluxEach       = r_RadUwFluxEach       + rs_RadUwFluxEach      (:,n)   * a_GQWeight(n)
        r_RadDwFluxEach       = r_RadDwFluxEach       + rs_RadDwFluxEach      (:,n)   * a_GQWeight(n)
        ra_DelRadUwFluxEach   = ra_DelRadUwFluxEach   + ras_DelRadUwFluxEach  (:,:,n) * a_GQWeight(n)
        ra_DelRadDwFluxEach   = ra_DelRadDwFluxEach   + ras_DelRadDwFluxEach  (:,:,n) * a_GQWeight(n)
        r_RadUwFluxSWEach     = r_RadUwFluxSWEach     + rs_RadUwFluxSWEach    (:,n)   * a_GQWeight(n)
        r_RadDwFluxSWEach     = r_RadDwFluxSWEach     + rs_RadDwFluxSWEach    (:,n)   * a_GQWeight(n)
        ra_DelRadUwFluxSWEach = ra_DelRadUwFluxSWEach + ras_DelRadUwFluxSWEach(:,:,n) * a_GQWeight(n)
        ra_DelRadDwFluxSWEach = ra_DelRadDwFluxSWEach + ras_DelRadDwFluxSWEach(:,:,n) * a_GQWeight(n)
      end do
      select case ( iCldClr )
      case ( 1 )
        FluxFactor = CloudFraction
      case ( 2 )
        FluxFactor = 1.0_DP - CloudFraction
      case default
        write( 6, * ) 'ERROR : Unexpected number of iCldClr, iCldClr'
        stop
      end select
      r_RadUwFlux       = r_RadUwFlux       + r_RadUwFluxEach       * FluxFactor
      r_RadDwFlux       = r_RadDwFlux       + r_RadDwFluxEach       * FluxFactor
      ra_DelRadUwFlux   = ra_DelRadUwFlux   + ra_DelRadUwFluxEach   * FluxFactor
      ra_DelRadDwFlux   = ra_DelRadDwFlux   + ra_DelRadDwFluxEach   * FluxFactor
      r_RadUwFluxLW     = r_RadUwFluxLW     + r_RadUwFluxLWEach     * FluxFactor
      r_RadDwFluxLW     = r_RadDwFluxLW     + r_RadDwFluxLWEach     * FluxFactor
      r_RadUwFluxSW     = r_RadUwFluxSW     + r_RadUwFluxSWEach     * FluxFactor
      r_RadDwFluxSW     = r_RadDwFluxSW     + r_RadDwFluxSWEach     * FluxFactor
      ra_DelRadUwFluxLW = ra_DelRadUwFluxLW + ra_DelRadUwFluxLWEach * FluxFactor
      ra_DelRadDwFluxLW = ra_DelRadDwFluxLW + ra_DelRadDwFluxLWEach * FluxFactor
      ra_DelRadUwFluxSW = ra_DelRadUwFluxSW + ra_DelRadUwFluxSWEach * FluxFactor
      ra_DelRadDwFluxSW = ra_DelRadDwFluxSW + ra_DelRadDwFluxSWEach * FluxFactor
    end do
    !
    if ( FlagOutSpeFlux ) then
      call RadFluxOutputFinalize
    end if



    call ConvCalcConvT2H( &
      & kmax, r_Press, r_Temp,  &
      & r_Enthalpy              &
      & )

    ! convection and add radiation by the use of enthalpy

    if ( FlagConvection ) then
      call ConvCalcRadConvTendency(                                   &
        & FlagMoistConvection,                                        & ! (in)
        & Grav, SurfHeatCap, DelTime,                                 & ! (in)
        & kmax, r_Press, r_Temp, SurfTemp, rm_VMR(:,1), r_MeanMolWt,  & ! (in)
        & r_Enthalpy,                                                 & ! (in)
        & NDelFlxMax,                                                 & ! (in)
        & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
        & r_DelEnthalpy, DelSurfTemp, r_DelVMRH2OVap, SensHeatFlux,   & ! (out)
        & FlagStdOut                                                  & ! (in)
        & )
    else
      call ConvCalcRadTendency(                                       &
        & Grav, SurfHeatCap, DelTime,                                 & ! (in)
        & kmax, r_Press, r_Temp,                                      & ! (in)
        & r_Enthalpy,                                                 & ! (in)
        & NDelFlxMax,                                                 & ! (in)
        & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
        & r_DelEnthalpy, DelSurfTemp                                  & ! (out)
        & )
      r_DelVMRH2OVap = 0.0d0
      SensHeatFlux   = 0.0d0
    end if
    r_Enthalpy = r_Enthalpy + r_DelEnthalpy
    !
    if ( FilterCoef > 0.0d0 ) then
      call FilterUsingEnthalpy( &
        & FilterCoef,    &
        & kmax, r_Press, &
        & r_Enthalpy     & ! (inout)
        & )
      ! Now, this is not calculated correctly.
      r_DelTempFilter = -1.0d100
    end if
    !
    call ConvCalcConvH2T(    &
      & kmax, r_Press, r_Enthalpy, &
      & r_TempTentative            &
      & )
    r_DelTempRC = r_TempTentative - r_Temp

    r_Temp      = r_Temp      + r_DelTempRC
    SurfTemp    = SurfTemp    + DelSurfTemp
    rm_VMR(:,1) = rm_VMR(:,1) + r_DelVMRH2OVap



    !-------------------------------------------------------
    ! for output

    r_RadUwFluxTentative = &
      &   r_RadUwFlux &
      & + ra_DelRadUwFlux(:,-1)*(SurfTemp -SurfTempB )
    do k = 0, NDelFlxMax
      r_RadUwFluxTentative = r_RadUwFluxTentative &
        & + ra_DelRadUwFlux(:,k)*(r_Temp(k)-r_TempB(k))
    end do
    r_RadDwFluxTentative = &
      &   r_RadDwFlux &
      & + ra_DelRadDwFlux(:,-1)*(SurfTemp -SurfTempB )
    do k = 0, NDelFlxMax
      r_RadDwFluxTentative = r_RadDwFluxTentative &
        & + ra_DelRadDwFlux(:,k)*(r_Temp(k)-r_TempB(k))
    end do

    if ( FlagStdOut ) then
      write( 6, * ) "Flux       : ", iTime, &
        &   r_RadUwFluxTentative(0)  &
        & - r_RadDwFluxTentative(0)  &
        & + SensHeatFlux,            &
        &   r_RadUwFluxTentative(0), &
        &   r_RadDwFluxTentative(0), &
        &   SensHeatFlux
      call flush( 6 )
    end if


    r_RadUwFluxLWTentative = &
      &   r_RadUwFluxLW &
      & + ra_DelRadUwFluxLW(:,-1)*(SurfTemp -SurfTempB )
    do k = 0, NDelFlxMax
      r_RadUwFluxLWTentative = r_RadUwFluxLWTentative &
        & + ra_DelRadUwFluxLW(:,k)*(r_Temp(k)-r_TempB(k))
    end do
    r_RadDwFluxLWTentative = &
      &   r_RadDwFluxLW &
      & + ra_DelRadDwFluxLW(:,-1)*(SurfTemp -SurfTempB )
    do k = 0, NDelFlxMax
      r_RadDwFluxLWTentative = r_RadDwFluxLWTentative &
        & + ra_DelRadDwFluxLW(:,k)*(r_Temp(k)-r_TempB(k))
    end do
    r_RadUwFluxSWTentative = &
      &   r_RadUwFluxSW &
      & + ra_DelRadUwFluxSW(:,-1)*(SurfTemp -SurfTempB )
    do k = 0, NDelFlxMax
      r_RadUwFluxSWTentative = r_RadUwFluxSWTentative &
        & + ra_DelRadUwFluxSW(:,k)*(r_Temp(k)-r_TempB(k))
    end do
    r_RadDwFluxSWTentative = &
      &   r_RadDwFluxSW &
      & + ra_DelRadDwFluxSW(:,-1)*(SurfTemp -SurfTempB )
    do k = 0, NDelFlxMax
      r_RadDwFluxSWTentative = r_RadDwFluxSWTentative &
        & + ra_DelRadDwFluxSW(:,k)*(r_Temp(k)-r_TempB(k))
    end do
    !
    ! The r_Rad*Flux*Tentative above will be used to calculate tendency and
    ! output below.
    !

    !-------------------------------------------------------



    ! CO2 condensation temperature (Pollack et al., 1981)
!!$    r_Temp = max( r_Temp, 149.2d0 + 6.48d0 * log( 0.135d0 * r_Press * 1.0d-2 ) )
!!$    SurfTemp = max( SurfTemp, 149.2d0 + 6.48d0 * log( 0.135d0 * r_Press(0) * 1.0d-2 ) )

    if ( FlagMajCompCond ) then

      write( 6, * ) 'ERROR : FlagMajCompCond has to be .false., now.'
      stop

      do k = 1, kmax
        z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
      end do

      m = IndexMajCompCloud
      !
      if ( FlagMajCompCloudEvolve ) then
        call CalcHeight( &
          & kmax,                                   &
          & r_Press, r_Temp, rm_VMR(:,1), r_MeanMolWt, &
          & r_Height                                &
          & )
        ! A routine below is commented out because it use constant Cp.
        ! (2020/07/24)
!!$        call CloudMajComp( &
!!$          & FlagMajCompCloudSedimentation, &
!!$          & FlagMajCompCloudSedimentationInfty, &
!!$          & Grav, Cp, MajCompLatentHeat, CloudNumRatio, DelTime, &
!!$          & kmax, r_Press, z_Press, r_Height, za_PtclDens(:,m), &
!!$          & za_PtclMMR(:,m), za_PtclEffRad(:,m), r_Temp &
!!$          & )

        if ( za_PtclMMR(1,m) > 0.0d0 ) then
          write( 6, * ) 'Atmosphere will collapse (EVOLVE mode): '
          do k = kmax, 0, -1
            TempSat = SatK1991CalcTempSat( r_Press(k) )
            write( 6, * ) '     ', k, r_Press(k), r_Temp(k), TempSat
          end do
          stop
        end if

      else
        DelMajCompCldMass = 0.0d0
        do k = kmax, 1, -1
          ! Sublimation

          ! A line below is commented out because it use constant Cp.
          ! (2020/07/24)
!!$          r_Temp(k) = r_Temp(k) &
!!$            & - DelMajCompCldMass * MajCompLatentHeat / ( Cp * z_DelAtmMass(k) )

          DelMajCompCldMass = 0.0d0
          !
          TempSat = SatK1991CalcTempSat( r_Press(k) )
          if ( r_Temp(k) <= TempSat ) then ! Condensation

            ! A line below is commented out because it use constant Cp.
            ! (2020/07/24)
!!$            DelMajCompCldMass = ( TempSat - r_Temp(k) ) * Cp * z_DelAtmMass(k) / MajCompLatentHeat

            r_Temp(k) = TempSat
            write( 6, * ) 'Atmosphere condenses at k = ', k, ' : ', r_Press(k), r_Temp(k), TempSat
          end if
        end do

        if ( DelMajCompCldMass > 0.0d0 ) then
          write( 6, * ) 'Atmosphere will collapse (non-EVOLVE mode): '
          do k = kmax, 0, -1
            TempSat = SatK1991CalcTempSat( r_Press(k) )
            write( 6, * ) '     ', k, r_Press(k), r_Temp(k), TempSat
          end do
          stop
        end if
      end if


!!$          za_PtclMMR   (k,m) = CO2IceMass / z_DelAtmMass(k)
!!$          ! M_ptcl = N_ptcl * 4/3*PI*R**3 * Rho
!!$          ! R**3 = M_ptcl / ( N_ptcl * 4/3*PI * Rho )
!!$          ! R = ( M_ptcl / ( N_ptcl * 4/3*PI * Rho ) )**(1/3)
!!$          ! R = ( M_ptcl / ( M_atm * Ratio * 4/3*PI * Rho ) )**(1/3)
!!$          ! 1.62e3 : CO2 density at 150 K (Mangan et al., 2017)
!!$          za_PtclEffRad(k,m) = &
!!$            & ( &
!!$            &     CO2IceMass &
!!$            &   / ( z_DelAtmMass(k) * CloudNumRatio * 4.0d0/3.0d0 * PI * 1.62e3 ) &
!!$            & )**(1.0d0/3.0d0) &
!!$            & + 1.0d-20  ! arbitrary lower limit

    end if


    if ( FlagAdjustAtmMass ) then


      ! Total mass adjustment
      !
      !   Estimate H2O mass change
      !

      do k = 1, kmax
        z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
      end do
      m = 1
      ! H2O
      !   mass mixing ratio
      rm_MMR(:,m) = rm_VMR(:,m) * m_HITRANMolWt(m_MolNum(m)) / r_MeanMolWt
      do k = 1, kmax
        zm_MMR(k,m) = ( rm_MMR(k-1,m) + rm_MMR(k,m) ) / 2.0d0
      end do
      !   mass
      m_AtmMassRadAct(m) = 0.0d0
      do k = 1, kmax
        m_AtmMassRadAct(m) = m_AtmMassRadAct(m) &
          & + z_DelAtmMass(k) * zm_MMR(k,m)
      end do
      !
      ! Total mass
      AtmMassTot = AtmMassDrySave + m_AtmMassRadAct(1)


      ! column mass of each component
      !   mean molecular weight
      call CalcMeanMolWt( &
        & kmax, NMolNum, m_MolNum, rm_VMR, r_VMRNonRadAct, &
        & r_MeanMolWt, &
        & .false. &
        & )
      ! mass mixing ratio
      !   non-radiatively active specie
      r_MMRNonRadAct = r_VMRNonRadAct * MolWtNonRadAct / r_MeanMolWt
      do k = 1, kmax
        z_MMRNonRadAct(k) = ( r_MMRNonRadAct(k-1) + r_MMRNonRadAct(k) ) / 2.0d0
      end do
      !   radiatively active specie
      do m = 1, NMolNum
        rm_MMR(:,m) = rm_VMR(:,m) * m_HITRANMolWt(m_MolNum(m)) / r_MeanMolWt
      end do
      do m = 1, NMolNum
        do k = 1, kmax
          zm_MMR(k,m) = ( rm_MMR(k-1,m) + rm_MMR(k,m) ) / 2.0d0
        end do
      end do
      ! column mass
      !   non-radiatively active specie
      AtmMassNonRadAct = 0.0d0
      do k = 1, kmax
        AtmMassNonRadAct = AtmMassNonRadAct &
          & + z_DelAtmMass(k) * z_MMRNonRadAct(k)
      end do
      !   radiatively active specie
      m_AtmMassRadAct = 0.0d0
      do m = 1, NMolNum
        do k = 1, kmax
          m_AtmMassRadAct(m) = m_AtmMassRadAct(m) &
            & + z_DelAtmMass(k) * zm_MMR(k,m)
        end do
      end do

      ! Pressure adjustment
      !
      r_Press(0) = AtmMassTot * Grav
      r_Press(1:kmax) = r_Press(0) * r_Sigma(1:kmax)
      do k = 1, kmax
        z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
      end do
      !
      do k = 1, kmax
        if ( r_Press(k-1) == 0.0d0 ) then
          z_Press(k) = ( r_Press(k-1) + r_Press(k) ) / 2.0d0
        else
          z_Press(k) = sqrt( r_Press(k-1) * r_Press(k) )
        end if
      end do

      ! Component mass adjustment
      !
!!$      r_VMRNonRadAct = r_VMRNonRadAct * AtmMassNonRadActSave / AtmMassNonRadAct
      if ( AtmMassNonRadAct /= 0.0d0 ) then
        r_VMRNonRadAct = r_VMRNonRadAct * AtmMassNonRadActSave / AtmMassNonRadAct
      else
        r_VMRNonRadAct = 0.0d0
      end if
      do m = 1+1, NMolNum
        ! H2O calculation is avoided
        if ( m_AtmMassRadAct(m) > 0.0d0 ) then
          rm_VMR(:,m) = rm_VMR(:,m) * m_AtmMassRadActSave(m) / m_AtmMassRadAct(m)
        end if
      end do
      rm_VMR = max( min( rm_VMR, 1.0d0 ), 0.0d0 )

!!$      write( 6, * ) 'Mass : ', AtmMassTot, AtmMassDrySave, AtmMassNonRadAct, m_AtmMassRadAct
    end if



    if ( mod( iTime, OutputStep ) == 0 ) then
      write( OutFN, '(a,i6.6,a)' ) &
        & trim( OutDir ) // "/" // OutNcFN(1:len_trim(OutNcFN)-3), iTime, '.nc'
      do m = 1, NPtcl
        a_PtclDens(m) = za_PtclDens(1,m)
      end do
      call CalcAdditionalVars(   &
        & Grav, kmax, r_Press, r_Temp, & ! (in)
        & r_Dens, r_AdLapRate          & ! (out)
        & )
      call IOProfileOutput( &
        & OutFN, &
        & kmax+1, NMolNum, r_Press, m_MolNum, &
        & r_Temp, rm_VMR, r_Press(0), SurfTemp, &
        & z_Press, &
        & NChar, NPtcl, a_PtclNum, a_PtclName, &
        & a_PtclDens, za_PtclEffRad, za_PtclMMR, &
        & r_Dens, r_AdLapRate                    &
        & )


      call CalcCp( &
        & kmax, r_Press, r_Temp,   & ! (in)
        & z_Cp                     & ! (out)
        & )
      call RadKDUtilsCalcTendencyVarCp( &
        & Grav, z_Cp, &
        & kmax, r_Press, &
        & r_RadUwFluxTentative, r_RadDwFluxTentative, &
        & z_DTempDt &
        & )
      call RadKDUtilsCalcTendencyVarCp( &
        & Grav, z_Cp, &
        & kmax, r_Press, &
        & r_RadUwFluxLWTentative, r_RadDwFluxLWTentative, &
        & z_DTempDtLW &
        & )
      call RadKDUtilsCalcTendencyVarCp( &
        & Grav, z_Cp, &
        & kmax, r_Press, &
        & r_RadUwFluxSWTentative, r_RadDwFluxSWTentative, &
        & z_DTempDtSW &
        & )
      call RadOutputFluxTendency( &
        & OutRadFluxNcFN, OutRadTendencyNcFN,   &
        & kmax, r_Press, z_Press, &
        & r_RadUwFluxTentative  , r_RadDwFluxTentative  , z_DTempDt  , &
        & r_RadUwFluxLWTentative, r_RadDwFluxLWTentative, z_DTempDtLW, &
        & r_RadUwFluxSWTentative, r_RadDwFluxSWTentative, z_DTempDtSW  &
        & )
      call OutputTendencies( &
        & OutTendenciesNcFN, &
        & DelTime, &
        & kmax, r_Press, &
        & r_DelTempRC, &
        & z_DTempDtLW, z_DTempDtSW, &
        & r_DelTempFilter,  &
        & DelSurfTemp  &
        & )

    end if


    AtmMassNonRadActSave = AtmMassNonRadAct
    m_AtmMassRadActSave  = m_AtmMassRadAct
    AtmMassDrySave       = AtmMassDry

  end do

  deallocate( a_GQMu     )
  deallocate( a_GQWeight )

  call RadFluxFinalize


  OutFN = trim( OutDir ) // "/" // OutNcFN
!!$  call Output( &
!!$    & OutFN, &
!!$    & kmax, r_Press, r_Temp, NMolNum, m_MolNum, rm_VMR &
!!$    & )
  do m = 1, NPtcl
    a_PtclDens(m) = za_PtclDens(1,m)
  end do
  call CalcAdditionalVars(   &
    & Grav, kmax, r_Press, r_Temp, & ! (in)
    & r_Dens, r_AdLapRate          & ! (out)
    & )
  call IOProfileOutput( &
    & OutFN, &
    & kmax+1, NMolNum, r_Press, m_MolNum, &
    & r_Temp, rm_VMR, r_Press(0), SurfTemp, &
    & z_Press, &
    & NChar, NPtcl, a_PtclNum, a_PtclName, &
    & a_PtclDens, za_PtclEffRad, za_PtclMMR, &
    & r_Dens, r_AdLapRate                    &
    & )


  deallocate( r_Press  )
  deallocate( m_MolNum )
  deallocate( r_Temp   )
  deallocate( rm_VMR   )

  deallocate( z_Press  )

  deallocate( r_Sigma )

  deallocate( a_PtclName     )
  deallocate( za_PtclDens    )
  deallocate( za_PtclEffRad  )
  deallocate( za_DelPtclMass )
  deallocate( za_DelPtclMassTentative )
  deallocate( za_PtclMMRforRad )

  deallocate( a_PtclDens     )

  deallocate( r_DelTempRC     )
  deallocate( r_DelTempFilter )
  deallocate( r_DelVMRH2OVap  )

  deallocate( r_Enthalpy      )
  deallocate( r_DelEnthalpy   )
  deallocate( r_TempTentative )

  deallocate( r_MeanMolWt     )
  deallocate( z_DTempDt       )
  deallocate( r_RadUwFlux     )
  deallocate( r_RadDwFlux     )
  deallocate( ra_DelRadUwFlux )
  deallocate( ra_DelRadDwFlux )

  deallocate( r_RadUwFluxLW     )
  deallocate( r_RadDwFluxLW     )
  deallocate( r_RadUwFluxSW     )
  deallocate( r_RadDwFluxSW     )
  deallocate( ra_DelRadUwFluxLW )
  deallocate( ra_DelRadDwFluxLW )
  deallocate( ra_DelRadUwFluxSW )
  deallocate( ra_DelRadDwFluxSW )

  deallocate( r_RadUwFluxEach       )
  deallocate( r_RadDwFluxEach       )
  deallocate( ra_DelRadUwFluxEach   )
  deallocate( ra_DelRadDwFluxEach   )
  deallocate( r_RadUwFluxLWEach     )
  deallocate( r_RadDwFluxLWEach     )
  deallocate( r_RadUwFluxSWEach     )
  deallocate( r_RadDwFluxSWEach     )
  deallocate( ra_DelRadUwFluxLWEach )
  deallocate( ra_DelRadDwFluxLWEach )
  deallocate( ra_DelRadUwFluxSWEach )
  deallocate( ra_DelRadDwFluxSWEach )

  deallocate( s_SZADeg               )
  deallocate( rs_RadUwFluxEach       )
  deallocate( rs_RadDwFluxEach       )
  deallocate( ras_DelRadUwFluxEach   )
  deallocate( ras_DelRadDwFluxEach   )
  deallocate( rs_RadUwFluxSWEach     )
  deallocate( rs_RadDwFluxSWEach     )
  deallocate( ras_DelRadUwFluxSWEach )
  deallocate( ras_DelRadDwFluxSWEach )
  !
!!$  deallocate( r_DelAtmMass )

  deallocate( a_CloudPtclNames )


  write( 6, * ) 'Finish'


end program main

!------------------------------------------------------------------------------

subroutine OutputTendencies( &
  & OutTendencyNcFN, &
  & DelTime, &
  & kmax, r_Press, &
  & r_DelTempRC, &
  & z_DTempDtLW, z_DTempDtSW, &
  & r_DelTempFilter,  &
  & DelSurfTemp  &
  & )

  use vtype_module

  use ni3_module

  implicit none

  character(*), intent(in ) :: OutTendencyNcFN
  real(DP)    , intent(in ) :: DelTime
  integer     , intent(in ) :: kmax
  real(DP)    , intent(in ) :: r_Press        (0:kmax)
  real(DP)    , intent(in ) :: r_DelTempRC    (0:kmax)
  real(DP)    , intent(in ) :: z_DTempDtLW    (1:kmax)
  real(DP)    , intent(in ) :: z_DTempDtSW    (1:kmax)
  real(DP)    , intent(in ) :: r_DelTempFilter(0:kmax)
  real(DP)    , intent(in ) :: DelSurfTemp


  ! Local variables
  !
  integer :: NcID
  character(extstr) :: Name
  character(extstr) :: StdName
  character(extstr) :: Units
  integer, parameter :: NDims1 = 1
  character(extstr)  :: a_DimName1(NDims1)
  character(extstr)  :: a_DimName0(0)

  real(DP), parameter :: MissingValue = -1.0d100

  real(DP)            :: r_DTempDtConv  (0:kmax)
  real(DP)            :: r_DTempDtSW    (0:kmax)
  real(DP)            :: r_DTempDtLW    (0:kmax)
  real(DP)            :: r_DTempDtFilter(0:kmax)
  real(DP)            :: DSurfTempDt


!!$  r_DTempDtSW(0     ) = 0.0d0
  r_DTempDtSW(0     ) = MissingValue
  r_DTempDtSW(1:kmax) = z_DTempDtSW
!!$  r_DTempDtLW(0     ) = 0.0d0
  r_DTempDtLW(0     ) = MissingValue
  r_DTempDtLW(1:kmax) = z_DTempDtLW

  r_DTempDtConv       = r_DelTempRC / DelTime - r_DTempDtSW - r_DTempDtLW
  r_DTempDtConv(0)    = MissingValue

  r_DTempDtFilter     = r_DelTempFilter / DelTime
  r_DTempDtFilter(0)  = MissingValue

  DSurfTempDt         = DelSurfTemp / DelTime


  call ni3_open( OutTendencyNcFN, "new", NcID )

  Name    = "Press"
  StdName = "plev"
  Units   = "Pa"
  call ni3_set_dim( NcID, name, NI3_DOUBLE, r_Press, &
    & stdname = stdname, units  = units )


  a_DimName1( 1 ) = "Press"

  Name    = "DTempDtSW"
  StdName = "radiative heating rate, short wave"
  Units   = "K s-1"
  call ni3_def_var( NcID, name, NI3_DOUBLE, NDims1, a_DimName1, &
    & stdname = stdname, units = units, fillvalue = MissingValue )
  call ni3_put_var( NcID, name, r_DTempDtSW )
  call ni3_put_att( NcID, name, "missing_value", MissingValue )

  Name    = "DTempDtLW"
  StdName = "radiative heating rate, long wave"
  Units   = "K s-1"
  call ni3_def_var( NcID, name, NI3_DOUBLE, NDims1, a_DimName1, &
    & stdname = stdname, units = units, fillvalue = MissingValue )
  call ni3_put_var( NcID, name, r_DTempDtLW )
  call ni3_put_att( NcID, name, "missing_value", MissingValue )

  Name    = "DTempDtConv"
  StdName = "convective heating rate"
  Units   = "K s-1"
  call ni3_def_var( NcID, name, NI3_DOUBLE, NDims1, a_DimName1, &
    & stdname = stdname, units = units, fillvalue = MissingValue )
  call ni3_put_var( NcID, name, r_DTempDtConv )
  call ni3_put_att( NcID, name, "missing_value", MissingValue )

!!$  Name    = "DTempDtFilter"
!!$  StdName = "heating rate by filter"
!!$  Units   = "K s-1"
!!$  call ni3_def_var( NcID, name, NI3_DOUBLE, NDims1, a_DimName1, &
!!$    & stdname = stdname, units = units, fillvalue = MissingValue )
!!$  call ni3_put_var( NcID, name, r_DTempDtFilter )
!!$  call ni3_put_att( NcID, name, "missing_value", MissingValue )

  Name    = "DSurfTempDt"
  StdName = "surface heating rate"
  Units   = "K s-1"
  call ni3_def_var( NcID, name, NI3_DOUBLE, 0, a_DimName0, &
    & stdname = stdname, units = units )
  call ni3_put_var( NcID, name, DSurfTempDt )
!!$  call ni3_put_att( NcID, name, "missing_value", MissingValue )

  call ni3_close( NcID )


end subroutine OutputTendencies

  !----------------------------------------------------------------------------

subroutine CalcCp( &
  & kmax, r_Press, r_Temp,   & ! (in)
  & z_Cp                     & ! (out)
  & )

  use vtype_module

#ifdef THERMOPROP_VIRACpLSF
  use thermo_prop_co2_viracplsf, only : &
#else
  use thermo_prop_co2_sw1996, only : &
#endif
    & ThermoPropDens,           &
    & ThermoPropCp

  integer , intent(in ) :: kmax
  real(dp), intent(in ) :: r_Press(0:kmax)
  real(dp), intent(in ) :: r_Temp (0:kmax)
  real(dp), intent(out) :: z_Cp   (1:kmax)


  ! Local variables
  !
  real(dp) :: Dens
  integer  :: k


  do k = 1, kmax
    Dens    = ThermoPropDens( r_Press(k), r_Temp(k) )
    z_Cp(k) = ThermoPropCp( Dens, r_Temp(k) )
  end do


end subroutine CalcCp

!------------------------------------------------------------------------------

subroutine CalcAdditionalVars(   &
  & Grav, kmax, r_Press, r_Temp, & ! (in)
  & r_Dens, r_AdLapRate          & ! (out)
  & )

  use vtype_module

#ifdef THERMOPROP_VIRACpLSF
  use thermo_prop_co2_viracplsf, only : &
#else
  use thermo_prop_co2_sw1996, only : &
#endif
    & ThermoPropDens, &
    & ThermoPropAdLapRateDTDP

  implicit none

  real(DP), intent(in ) :: Grav
  integer , intent(in ) :: kmax
  real(DP), intent(in ) :: r_Press    (0:kmax)
  real(DP), intent(in ) :: r_Temp     (0:kmax)
  real(DP), intent(out) :: r_Dens     (0:kmax)
  real(DP), intent(out) :: r_AdLapRate(0:kmax)

  ! Local variables
  !
  integer  :: k


  ! dT/dz = dT/dp dp/dz = dT/dp (-rho g)
  do k = 0, kmax
    r_Dens(k) = ThermoPropDens( r_Press(k), r_Temp(k) )
    r_AdLapRate(k) = &
      & - ThermoPropAdLapRateDTDP( r_Dens(k), r_Temp(k) ) &
      &   * ( - r_Dens(k) * Grav )
  end do


end subroutine CalcAdditionalVars

!------------------------------------------------------------------------------

subroutine CalcMeanMolWt( &
  & kmax, NMolNum, m_MolNum, rm_VMR, r_VMRNonRadAct, &
  & r_MeanMolWt, &
  & FlagOutput &
  & )

  use vtype_module

  use rad_planetconst, only : MolWtNonRadAct
  use hitranconst, only : m_HITRANMolWt

  implicit none

  integer , intent(in ) :: kmax
  integer , intent(in ) :: NMolNum
  integer , intent(in ) :: m_MolNum      (1:NMolNum)
  real(DP), intent(in ) :: rm_VMR        (0:kmax,1:NMolNum)
  real(DP), intent(in ) :: r_VMRNonRadAct(0:kmax)
  real(DP), intent(out) :: r_MeanMolWt   (0:kmax)
  logical , intent(in ) :: FlagOutput


  ! Local variables
  !
  real(dp) :: r_SumVMR(0:kmax)

  integer  :: k
  integer  :: m


  r_MeanMolWt = 0.0d0
  r_MeanMolWt = r_MeanMolWt + r_VMRNonRadAct * MolWtNonRadAct
  r_SumVMR = 0.0d0
  r_SumVMR = r_SumVMR + r_VMRNonRadAct
  do m = 1, NMolNum
    r_MeanMolWt = r_MeanMolWt + rm_VMR(:,m) * m_HITRANMolWt(m_MolNum(m))
    r_SumVMR    = r_SumVMR    + rm_VMR(:,m)
  end do
  ! Now, this is not normalized.
!!$  r_MeanMolWt = r_MeanMolWt / r_SumVMR


  if ( FlagOutput ) then
    write( 6, * ) 'Mean molecular weight'
    do k = 0, kmax
      write( 6, * ) k, r_MeanMolWt(k)
    end do
  end if


end subroutine CalcMeanMolWt

!------------------------------------------------------------------------------
!!$
!!$subroutine ConvectionTrial( &
!!$  & kmax, NMolNum, r_Press, r_MeanMolWt, &
!!$  & r_Temp, rm_VMR &
!!$  & )
!!$
!!$  use vtype_module
!!$  use constants0, only : GasRUniv
!!$  use rad_planetconst, only : &
!!$    & Cp,       &
!!$    & MolWtH2O, &
!!$    & CpH2OVap
!!$
!!$  implicit none
!!$
!!$  integer , intent(in   ) :: kmax
!!$  integer , intent(in   ) :: NMolNum
!!$  real(DP), intent(in   ) :: r_Press(0:kmax)
!!$  real(DP), intent(in   ) :: r_MeanMolWt(0:kmax)
!!$  real(DP), intent(inout) :: r_Temp     (0:kmax)
!!$  real(DP), intent(inout) :: rm_VMR     (0:kmax,1:NMolNum)
!!$
!!$
!!$  ! Local variables
!!$  !
!!$  real(DP), parameter :: LatentHeat = 2.5d6    ! J kg-1
!!$
!!$  real(DP) :: CPPerMol
!!$  real(DP) :: CpH2OVapPerMol
!!$  real(DP) :: CPTotPerMol
!!$  real(DP) :: LatentHeatPerMol
!!$  real(DP) :: PressH2OVapSat
!!$  real(DP) :: DPressH2OVapSatDTemp
!!$  real(DP) :: VMRSat
!!$  real(DP) :: VMRSave
!!$  real(DP) :: DelTemp
!!$  real(DP) :: TempTmp
!!$
!!$  real(DP) :: LOCALPressH2OVapSat
!!$  real(DP) :: LOCALDPressH2OVapSatDTemp
!!$
!!$  logical  :: FlagMoistConvection
!!$  logical  :: FlagConnectToSurface
!!$
!!$  integer  :: k
!!$
!!$
!!$  FlagMoistConvection  = .true.
!!$  FlagConnectToSurface = .true.
!!$
!!$
!!$  LatentHeatPerMol = LatentHeat * MolWtH2O
!!$  CpH2OVapPerMol   = CpH2OVap * MolWtH2O
!!$
!!$  PressH2OVapSat   = LOCALPressH2OVapSat( LatentHeatPerMol, r_Temp(0) )
!!$  VMRSat           = PressH2OVapSat / r_Press(0)
!!$  rm_VMR(0,1)      = VMRSat
!!$
!!$  VMRSave = rm_VMR(0,1)
!!$
!!$  do k = 1, kmax
!!$
!!$    if ( FlagMoistConvection ) then
!!$      ! moist convection
!!$
!!$      CpPerMol = Cp * r_MeanMolWt(k-1)
!!$      PressH2OVapSat = LOCALPressH2OVapSat( LatentHeatPerMol, r_Temp(k-1) )
!!$      DPressH2OVapSatDTemp = LOCALDPressH2OVapSatDTemp( LatentHeatPerMol, r_Temp(k-1) )
!!$      VMRSat = PressH2OVapSat / r_Press(k-1)
!!$      CpTotPerMol = CpPerMol * ( 1.0d0 - VMRSat ) + CpH2OVapPerMol * VMRSat
!!$      DelTemp = ( GasRUniv * r_Temp(k-1) + LatentHeatPerMol * VMRSat ) &
!!$        & / ( CPTotPerMol + LatentHeatPerMol / r_Press(k-1) * DPressH2OVapSatDTemp ) &
!!$        & * log( r_Press(k) / r_Press(k-1) )
!!$      !
!!$      if ( FlagConnectToSurface ) then
!!$        TempTmp = r_Temp(k-1) + DelTemp
!!$        if ( r_Temp(k) < TempTmp ) then
!!$          r_Temp(k)     = TempTmp
!!$          rm_VMR(k-1,1) = VMRSat
!!$        else
!!$          FlagConnectToSurface = .false.
!!$        end if
!!$      end if
!!$      if ( .not. FlagConnectToSurface ) then
!!$        rm_VMR(k-1,1) = VMRSave
!!$        if ( rm_VMR(k-1,1) > VMRSat ) then
!!$          rm_VMR(k-1,1) = VMRSat
!!$        end if
!!$        VMRSave = rm_VMR(k-1,1)
!!$      end if
!!$
!!$    else
!!$      ! dry convection
!!$!      GasConst = GasRUniv / r_MeanMolWt(k-1)
!!$!      DelTemp = GasConst * TempLL / Cp * log( z_Press(k) / PressLL )
!!$      CpPerMol = Cp * r_MeanMolWt(k-1)
!!$      DelTemp = GasRUniv * r_Temp(k-1) / CpPerMol * log( r_Press(k) / r_Press(k-1) )
!!$      if ( FlagConnectToSurface ) then
!!$        TempTmp = r_Temp(k-1) + DelTemp
!!$        if ( r_Temp(k) < TempTmp ) then
!!$          r_Temp(k)     = TempTmp
!!$        else
!!$          FlagConnectToSurface = .false.
!!$        end if
!!$      end if
!!$    end if
!!$    !
!!$    !
!!$    if ( k == kmax ) then
!!$      rm_VMR(k,1) = rm_VMR(k-1,1)
!!$    end if
!!$  end do
!!$
!!$
!!$end subroutine ConvectionTrial

!------------------------------------------------------------------------------

!!$function LOCALPressH2OVapSat( LatentHeatPerMol, Temp ) result( PressH2OVapSat )
!!$
!!$  use vtype_module
!!$  use constants0, only : GasRUniv
!!$
!!$  real(DP), intent(in ) :: LatentHeatPerMol
!!$  real(DP), intent(in ) :: Temp
!!$
!!$  real(DP) :: PressH2OVapSat
!!$
!!$
!!$  PressH2OVapSat = 1.4d11 * exp( - LatentHeatPerMol / ( GasRUniv * Temp ) )
!!$
!!$end function LOCALPressH2OVapSat

!------------------------------------------------------------------------------

!!$function LOCALDPressH2OVapSatDTemp( LatentHeatPerMol, Temp ) result( DPressH2OVapSatDTemp )
!!$
!!$  use vtype_module
!!$  use constants0, only : GasRUniv
!!$
!!$  real(DP), intent(in ) :: LatentHeatPerMol
!!$  real(DP), intent(in ) :: Temp
!!$
!!$  real(DP) :: DPressH2OVapSatDTemp
!!$
!!$
!!$  DPressH2OVapSatDTemp = &
!!$    & 1.4d11 * exp( - LatentHeatPerMol / ( GasRUniv * Temp ) ) &
!!$    & * LatentHeatPerMol / ( GasRUniv * Temp**2 )
!!$
!!$end function LOCALDPressH2OVapSatDTemp

!------------------------------------------------------------------------------

subroutine Diffusion( &
  & DelTime, &
  & kmax, r_Press, z_Press, r_MeanMolWt, &
  & z_Temp &
  & )

  use vtype_module
  use constants0, only : GasRUniv
  use rad_planetconst, only : Grav

  implicit none

  real(DP), intent(in   ) :: DelTime
  integer , intent(in   ) :: kmax
  real(DP), intent(in   ) :: r_Press(0:kmax)
  real(DP), intent(in   ) :: z_Press(1:kmax)
  real(DP), intent(in   ) :: r_MeanMolWt(0:kmax)
  real(DP), intent(inout) :: z_Temp     (1:kmax)


  ! Local variables
  !
  real(DP) :: DiffCoef = 0.0d0
!!$  real(DP) :: DiffCoef = 1.0d0
  real(DP) :: r_Flux(0:kmax)
  real(DP) :: GasConst
  real(DP) :: Temp
  real(DP) :: Rho

  integer  :: k


  k = 0
  r_Flux(k) = 0.0d0
  do k = 0+1, kmax-1
    GasConst = GasRUniv / r_MeanMolWt(k)
    Temp = ( z_Temp(k) + z_Temp(k+1) ) / 2.0d0
    Rho = r_Press(k) / ( GasConst * Temp )
    r_Flux(k) = &
      &   DiffCoef &
      & * ( z_Temp(k+1) - z_Temp(k) ) / ( z_Press(k+1) - z_Press(k) ) &
      & * Rho**2
  end do
  k = kmax
  r_Flux(k) = 0.0d0


  do k = 1, kmax
    z_Temp(k) = z_Temp(k) &
      & + Grav * ( r_Flux(k) - r_Flux(k-1) ) / ( r_Press(k) - r_Press(k-1) ) &
      &   * DelTime
  end do


end subroutine Diffusion
