module particle_ec_merge

  use vtype_module

  implicit none


  public :: ParticleECMergeWaveNumber

  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------

  subroutine ParticleECMergeWaveNumber( &
    & NProcs, InputFileName0, OutputFileName &
    & )

    use ni3_module

    integer         , intent(in) :: NProcs
    character(len=*), intent(in) :: InputFileName0
    character(len=*), intent(in) :: OutputFileName


    !
    ! Local variables
    !
    integer                        :: InNcID
    integer                        :: OutNcID
    character(stdstr)              :: Mode
    integer                        :: NDims
    character(stdstr), allocatable :: a_DimNames( : )
    character(stdstr)              :: Name

    integer                   :: NMolNum
    integer                   :: NCharNum
    integer                   :: NPtclNum
    integer                   :: NPtclRadiusNum

    character(len=extstr)     :: InputFileName

    integer     , allocatable :: m_MolNum       ( : )
    integer     , allocatable :: a_CharNum      ( : )
    integer     , allocatable :: a_PtclNum      ( : )
    integer     , allocatable :: r_PtclRadiusNum( : )

    integer          , allocatable :: a_PtclNRadius  ( : )
    character(stdstr), allocatable :: a_PtclName   ( : )
    real(dp)         , allocatable :: ra_PtclRadius( :, : )

    real(dp)                  :: RayScatCoefNonRadAct
    real(dp)    , allocatable :: wm_RayScatCoef(:,:)
    real(dp)    , allocatable :: wra_QExt      (:,:,:)
    real(dp)    , allocatable :: wra_SSA       (:,:,:)
    real(dp)    , allocatable :: wra_AF        (:,:,:)

    real(dp)                  :: WaveNum


    integer          , parameter :: ndims1 = 1
    character(stdstr)            :: dname1( ndims1 )
    integer          , parameter :: ndims2 = 2
    character(stdstr)            :: dname2( ndims2 )
    integer          , parameter :: ndims3 = 3
    character(stdstr)            :: dname3( ndims3 )

    integer :: St2(2)
    integer :: Co2(2)
    integer :: St3(3)
    integer :: Co3(3)

    integer :: iRank
    integer :: NWaveNum
    integer :: NWaveNumEach

    integer :: iWaveNum

    integer :: l


    NWaveNum = 0

    do iRank = 0, NProcs-1

      write( InputFileName, '(a,a,i6.6,a)' ) &
        & InputFileName0(1:len_trim(InputFileName0)-3), &
        & "_", &
        & iRank, &
        & ".nc"

      Mode = "read"
      call ni3_open( InputFileName, Mode, InNcID )

      if ( iRank == 0 ) then
        call ni3_inq_dimlen( InNcID, "MolNum"       , NMolNum        )
        call ni3_inq_dimlen( InNcID, "CharNum"      , NCharNum       )
        call ni3_inq_dimlen( InNcID, "PtclNum"      , NPtclNum       )
        call ni3_inq_dimlen( InNcID, "PtclRadiusNum", NPtclRadiusNum )

        allocate( m_MolNum       ( NMolNum        ) )
        allocate( a_CharNum      ( NCharNum       ) )
        allocate( a_PtclNum      ( NPtclNum       ) )
        allocate( r_PtclRadiusNum( NPtclRadiusNum ) )
        allocate( a_PtclNRadius( NPtclNum ) )
        allocate( a_PtclName   ( NPtclNum ) )
        allocate( ra_PtclRadius( NPtclRadiusNum, NPtclNum ) )

        call ni3_get_var( InNcID, "MolNum"       , m_MolNum        )
        call ni3_get_var( InNcID, "CharNum"      , a_CharNum       )
        call ni3_get_var( InNcID, "PtclNum"      , a_PtclNum       )
        call ni3_get_var( InNcID, "PtclRadiusNum", r_PtclRadiusNum )
        call ni3_get_var( InNcID, "PtclNRadius"  , a_PtclNRadius )
        call ni3_get_var( InNcID, "PtclName"     , a_PtclName )
        call ni3_get_var( InNcID, "PtclRadius"   , ra_PtclRadius )

      end if

      call ni3_inq_dimlen( InNcID, "WaveNum", NWaveNumEach )
      NWaveNum = NWaveNum + NWaveNumEach

      call ni3_close( InNcID )

    end do


    allocate( wm_RayScatCoef( 1, NMolNum ) )
    allocate( wra_QExt      ( 1, NPtclRadiusNum, NPtclNum ) )
    allocate( wra_SSA       ( 1, NPtclRadiusNum, NPtclNum ) )
    allocate( wra_AF        ( 1, NPtclRadiusNum, NPtclNum ) )


    Mode = "new"
    call ni3_open( OutputFileName, Mode, OutNcID )

    iWaveNum = 0

    do iRank = 0, NProcs-1

      write( InputFileName, '(a,a,i6.6,a)' ) &
        & InputFileName0(1:len_trim(InputFileName0)-3), &
        & "_", &
        & iRank, &
        & ".nc"

      Mode = "read"
      call ni3_open( InputFileName, Mode, InNcID )


      if ( iRank == 0 ) then
        Name = "_GLOBAL_"
        call ni3_cp_atts( InNcID, OutNcID, Name )

        Name = "WaveNum"
!!$        call ni3_def_dim( OutNcID, Name, NI3_DOUBLE, NI3_UNLIMITED )
        call ni3_def_dim( OutNcID, Name, NI3_DOUBLE, NWaveNum )
        call ni3_cp_atts( InNcID, OutNcID, Name )

        Name = "MolNum"
        call ni3_set_dim( OutNcID, Name, NI3_INT, m_MolNum )
        call ni3_cp_atts( InNcID, OutNcID, Name )

        if ( NPtclNum > 0 ) then
          Name = "CharNum"
          call ni3_set_dim( OutNcID, Name, NI3_INT, a_CharNum )
          call ni3_cp_atts( InNcID, OutNcID, Name )

          Name = "PtclNum"
          call ni3_set_dim( OutNcID, Name, NI3_INT, a_PtclNum )
          call ni3_cp_atts( InNcID, OutNcID, Name )

          Name = "PtclRadiusNum"
          call ni3_set_dim( OutNcID, Name, NI3_INT, r_PtclRadiusNum )
          call ni3_cp_atts( InNcID, OutNcID, Name )

          dname1( 1 ) = "PtclNum"
          !
          name    = "PtclNRadius"
          call ni3_def_var( OutNcID, name, NI3_INT, ndims1, dname1 )
          call ni3_put_var( OutNcID, name, a_PtclNRadius )
          call ni3_cp_atts( InNcID, OutNcID, Name )
          !
          dname2( 1 ) = "CharNum"
          dname2( 2 ) = "PtclNum"
          !
          name    = "PtclName"
          call ni3_def_var( OutNcID, name, NI3_CHAR, ndims2, dname2 )
          call ni3_put_var( OutNcID, name, a_PtclName )
          call ni3_cp_atts( InNcID, OutNcID, Name )
          !
          dname2( 1 ) = "PtclRadiusNum"
          dname2( 2 ) = "PtclNum"
          !
          name    = "PtclRadius"
          call ni3_def_var( OutNcID, name, NI3_DOUBLE, ndims2, dname2 )
          call ni3_put_var( OutNcID, name, ra_PtclRadius )
          call ni3_cp_atts( InNcID, OutNcID, Name )
        end if


        dname1( 1 ) = "WaveNum"
        !
        name    = "RayScatCoefNonRadAct"
        call ni3_def_var( OutNcID, name, NI3_DOUBLE, ndims1, dname1 )
        call ni3_cp_atts( InNcID, OutNcID, Name )

        dname2( 1 ) = "WaveNum"
        dname2( 2 ) = "MolNum"
        !
        name    = "RayScatCoef"
        call ni3_def_var( OutNcID, name, NI3_DOUBLE, ndims2, dname2 )
        call ni3_cp_atts( InNcID, OutNcID, Name )

        if ( NPtclNum > 0 ) then
          dname3( 1 ) = "WaveNum"
          dname3( 2 ) = "PtclRadiusNum"
          dname3( 3 ) = "PtclNum"
          !
          name    = "QExt"
          call ni3_def_var( OutNcID, name, NI3_DOUBLE, ndims3, dname3 )
          call ni3_cp_atts( InNcID, OutNcID, Name )
          !
          name    = "SSA"
          call ni3_def_var( OutNcID, name, NI3_DOUBLE, ndims3, dname3 )
          call ni3_cp_atts( InNcID, OutNcID, Name )
          !
          name    = "AF"
          call ni3_def_var( OutNcID, name, NI3_DOUBLE, ndims3, dname3 )
          call ni3_cp_atts( InNcID, OutNcID, Name )
        end if

      end if


      Name = "WaveNum"
      call ni3_inq_dimlen( InNcID, Name, NWaveNumEach )

      do l = 1, NWaveNumEach
        iWaveNum = iWaveNum + 1

        Name = "WaveNum"
        call ni3_get_varss( InNcID, Name, l, WaveNum )
        call ni3_put_varss( OutNcID, Name, iWaveNum, WaveNum )

        Name = "RayScatCoefNonRadAct"
        call ni3_get_varss( InNcID, Name, l, RayScatCoefNonRadAct )
        call ni3_put_varss( OutNcID, Name, iWaveNum, RayScatCoefNonRadAct )

        Name = "RayScatCoef"
        Co2(1) = 1
        Co2(2) = NMolNum
        St2(1) = l
        St2(2) = 1
        call ni3_get_varpart( InNcID , Name, St2, Co2, wm_RayScatCoef )
        St2(1) = iWaveNum
        St2(2) = 1
        call ni3_put_varpart( OutNcID, Name, St2, Co2, wm_RayScatCoef )

        if ( NPtclNum > 0 ) then

          Co3(1) = 1
          Co3(2) = NPtclRadiusNum
          Co3(3) = NPtclNum
          St3(1) = l
          St3(2) = 1
          St3(3) = 1
          Name = "QExt"
          call ni3_get_varpart( InNcID , Name, St3, Co3, wra_QExt )
          Name = "SSA"
          call ni3_get_varpart( InNcID , Name, St3, Co3, wra_SSA  )
          Name = "AF"
          call ni3_get_varpart( InNcID , Name, St3, Co3, wra_AF   )
          St3(1) = iWaveNum
          St3(2) = 1
          St3(3) = 1
          Name = "QExt"
          call ni3_put_varpart( OutNcID, Name, St3, Co3, wra_QExt )
          Name = "SSA"
          call ni3_put_varpart( OutNcID, Name, St3, Co3, wra_SSA  )
          Name = "AF"
          call ni3_put_varpart( OutNcID, Name, St3, Co3, wra_AF   )
        end if

      end do

      call ni3_close( InNcID )

    end do

    call ni3_close( OutNcID )


    deallocate( wm_RayScatCoef )
    deallocate( wra_QExt       )
    deallocate( wra_SSA        )
    deallocate( wra_AF         )


  end subroutine ParticleECMergeWaveNumber

  !----------------------------------------------------------------------------

end module particle_ec_merge
