module rad_output

  use vtype_module

  implicit none

  private

  public :: RadOutputFluxTendency

contains

  !----------------------------------------------------------------------------

  subroutine RadOutputFluxTendency( &
    & OutFluxNcFN, OutTendencyNcFN, &
    & kmax, r_Press, z_Press, &
    & r_RadUwFlux, r_RadDwFlux, z_DTempDt, &
    & r_RadUwFluxLW, r_RadDwFluxLW, z_DTempDtLW, &
    & r_RadUwFluxSW, r_RadDwFluxSW, z_DTempDtSW, &
    & WNStart, WNEnd &
    & )

    use ni3_module
    use netcdf

    character(*), intent(in ) :: OutFluxNcFN
    character(*), intent(in ) :: OutTendencyNcFN
    integer     , intent(in ) :: kmax
    real(DP)    , intent(in ) :: r_Press    (0:kmax)
    real(DP)    , intent(in ) :: z_Press    (1:kmax)
    real(DP)    , intent(in ) :: r_RadUwFlux(0:kmax)
    real(DP)    , intent(in ) :: r_RadDwFlux(0:kmax)
    real(DP)    , intent(in ) :: z_DTempDt  (1:kmax)
    real(DP)    , intent(in ), optional :: r_RadUwFluxLW(0:kmax)
    real(DP)    , intent(in ), optional :: r_RadDwFluxLW(0:kmax)
    real(DP)    , intent(in ), optional :: z_DTempDtLW  (1:kmax)
    real(DP)    , intent(in ), optional :: r_RadUwFluxSW(0:kmax)
    real(DP)    , intent(in ), optional :: r_RadDwFluxSW(0:kmax)
    real(DP)    , intent(in ), optional :: z_DTempDtSW  (1:kmax)
    real(DP)    , intent(in ), optional :: WNStart
    real(DP)    , intent(in ), optional :: WNEnd


    ! Local variables
    !
    integer :: NcID
    character(extstr) :: Name
    character(extstr) :: StdName
    character(extstr) :: LongName
    character(extstr) :: Units
    integer, parameter :: NDims1 = 1
    character(extstr)  :: a_DimName1(NDims1)


    call ni3_open( OutFluxNcFN, "new", NcID )

    Name     = "Press"
    StdName  = "plev"
    LongName = "pressure"
    Units    = "Pa"
    call ni3_set_dim( NcID, name, NF90_DOUBLE, r_Press, &
      & stdname = stdname, longname = longname, units  = units )
    if ( present( WNStart ) .and. present( WNEnd ) ) then
      call ni3_put_att( NcID, "_global_", "wn_start", WNStart )
      call ni3_put_att( NcID, "_global_", "wn_end"  , WNEnd   )
      call ni3_put_att( NcID, "_global_", "wn_unit" , "m-1"   )
    end if

    a_DimName1( 1 ) = "Press"

    Name    = "RadUwFlux"
    StdName = "upward radiation flux"
    Units   = "W m-2"
    call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
      stdname = stdname, units = units )
    call ni3_put_var( NcID, name, r_RadUwFlux )

    Name    = "RadDwFlux"
    StdName = "upward radiation flux"
    Units   = "W m-2"
    call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
      stdname = stdname, units = units )
    call ni3_put_var( NcID, name, r_RadDwFlux )

    if ( present( r_RadUwFluxLW ) ) then
      Name     = "RadUwFluxLW"
      StdName  = "upward longwave radiation flux"
      LongName = StdName
      Units    = "W m-2"
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
        stdname = StdName, longname = LongName, units = Units )
      call ni3_put_var( NcID, name, r_RadUwFluxLW )
    end if

    if ( present( r_RadDwFluxLW ) ) then
      Name     = "RadDwFluxLW"
      StdName  = "downward longwave radiation flux"
      LongName = StdName
      Units    = "W m-2"
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
        stdname = StdName, longname = LongName, units = Units )
      call ni3_put_var( NcID, name, r_RadDwFluxLW )
    end if

    if ( present( r_RadUwFluxSW ) ) then
      Name     = "RadUwFluxSW"
      StdName  = "upward shortwave radiation flux"
      LongName = StdName
      Units    = "W m-2"
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
        stdname = StdName, longname = LongName, units = Units )
      call ni3_put_var( NcID, name, r_RadUwFluxSW )
    end if

    if ( present( r_RadDwFluxSW ) ) then
      Name     = "RadDwFluxSW"
      StdName  = "downward shortwave radiation flux"
      LongName = StdName
      Units    = "W m-2"
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
        stdname = StdName, longname = LongName, units = Units )
      call ni3_put_var( NcID, name, r_RadDwFluxSW )
    end if


    call ni3_close( NcID )



    call ni3_open( OutTendencyNcFN, "new", NcID )

    Name     = "Press"
    StdName  = "plev"
    LongName = "pressure"
    Units    = "Pa"
    call ni3_set_dim( NcID, name, NF90_DOUBLE, z_Press, &
      & stdname = stdname, longname = longname, units  = units )


    a_DimName1( 1 ) = "Press"

    Name    = "DTempDt"
    StdName = "radiative heating rate"
    Units   = "K s-1"
    call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
      stdname = stdname, units = units )
    call ni3_put_var( NcID, name, z_DTempDt )

    if ( present( z_DTempDtLW ) ) then
      Name     = "DTempDtLW"
      StdName  = "longwave radiative heating rate"
      LongName = StdName
      Units    = "K s-1"
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
        stdname = StdName, longname = LongName, units = Units )
      call ni3_put_var( NcID, name, z_DTempDtLW )
    end if

    if ( present( z_DTempDtSW ) ) then
      Name     = "DTempDtSW"
      StdName  = "shortwave radiative heating rate"
      LongName = StdName
      Units    = "K s-1"
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims1, a_DimName1, &
        stdname = StdName, longname = LongName, units = Units )
      call ni3_put_var( NcID, name, z_DTempDtSW )
    end if

    call ni3_close( NcID )


  end subroutine RadOutputFluxTendency

  !----------------------------------------------------------------------------

end module rad_output
