module ac_calc_utils

  use vtype_module

  implicit none

  private

  public :: ACCalcUtilsModifyWavenumberRange1
  public :: ACCalcUtilsModifyWavenumberRange2
  public :: ACCalcUtilsModifyWavenumberRange3

  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------

  subroutine ACCalcUtilsModifyWavenumberRange1( &
    & NProcs, MyRank, &
    & DelWaveNum, &
    & WaveNumS, WaveNumE &
    & )

    integer , intent(in   ) :: NProcs
    integer , intent(in   ) :: MyRank
    real(dp), intent(in   ) :: DelWaveNum
    real(dp), intent(inout) :: WaveNumS
    real(dp), intent(inout) :: WaveNumE


    ! Local variables
    !
    integer :: NWaveNum
    integer :: iRank


    NWaveNum = int( ( WaveNumE - WaveNumS ) / DelWaveNum ) + 1
    WaveNumE = WaveNumS - DelWaveNum
    do iRank = 0, MyRank
      WaveNumS = WaveNumE + DelWaveNum
      WaveNumE = WaveNumS + ( int( NWaveNum / NProcs ) - 1 ) * DelWaveNum
      if ( iRank+1 <= mod( NWaveNum, NProcs ) ) then
        WaveNumE = WaveNumE + DelWaveNum
      end if
    end do


  end subroutine ACCalcUtilsModifyWavenumberRange1

  !----------------------------------------------------------------------------

  subroutine ACCalcUtilsModifyWavenumberRange2( &
    & FlagNetCDF, NHIT, a_HITFileName, &
    & NProcs, MyRank, &
    & NMolNum, m_MolNum, &
    & DelWaveNum, &
    & WaveNumS, WaveNumE &
    & )

    use lbl_module

    logical         , intent(in   ) :: FlagNetCDF
    integer         , intent(in   ) :: NHIT
    character(len=*), intent(in   ) :: a_HITFileName( NHIT )
    integer         , intent(in   ) :: NProcs
    integer         , intent(in   ) :: MyRank
    integer         , intent(in   ) :: NMolNum
    integer         , intent(in   ) :: m_MolNum(NMolNum)
    real(dp)        , intent(in   ) :: DelWaveNum
    real(dp)        , intent(inout) :: WaveNumS
    real(dp)        , intent(inout) :: WaveNumE


    ! Local variables
    !
    integer, parameter :: MolNumChk = 2
    logical            :: FlagClear

    type( LineParam )         :: LParam
    character(len=extstr)     :: FileName

    integer  :: iLineS
    integer  :: iLineE
    integer  :: NLine
    integer  :: iNum
    real(dp) :: an_WaveNumBnds(2,0:NProcs-1)

    integer :: l
    integer :: m
    integer :: n


    FlagClear = .false.
    do m = 1, NMolNum
      if ( m_MolNum(m) == MolNumChk ) then
        FlagClear = .true.
      end if
    end do
    if ( .not. FlagClear ) then
      write( 6, * ) 'Unable to find ', MolNumChk, 'th molecule.'
      write( 6, * ) ' ', m_MolNum
      stop
    end if


    m = MolNumChk

    FileName = a_HITFileName( m )
    write( 6, '(a)' ) '  Line File: ' // trim( FileName )

    if ( FlagNetCDF ) then
      call lbl_readfile_netcdf( FileName, 1, LParam )
      write( 6, '(a,i9)' ) '             Number of lines : ', LParam%NLine
    else
      write( 6, * ) 'ASCII format HITRAN data file is not supported, now.'
      stop
    end if


    do l = 1, LParam%NLine
      if ( WaveNumS <= LParam%wnlc0(l) ) exit
    end do
    iLineS = l
    iLineS = min( max( iLineS, 1 ), LParam%NLine )
    do l = iLineS, LParam%NLine
      if ( WaveNumE < LParam%wnlc0(l) ) exit
    end do
    iLineE = l-1
    iLineE = min( max( iLineE, iLineS ), LParam%NLine )

    NLine = iLineE - iLineS + 1

    if ( NLine < NProcs ) then
      write( 6, * ) 'Number of processes are too large.'
      write( 6, * ) '  ', NLine, NProcs
      stop
    end if

    do n = 0, NProcs-1
      if ( n == 0 ) then
        an_WaveNumBnds(1,n) = WaveNumS
      else
!!$        a_WaveNumBnds(1,n) = a_WaveNumBnds(2,n-1) + DelWaveNum
        an_WaveNumBnds(1,n) = WaveNumS + (iNum+1) * DelWaveNum
      end if
      if ( n == (NProcs-1) ) then
        an_WaveNumBnds(2,n) = WaveNumE
      else
        l = iLineS + int( real( NLine ) / NProcs * ( n+1 ) )
        l = min( l, LParam%NLine )

        ! restrict iNum
        if ( ( iNum + ( (NProcs-1) - n ) ) > int( (WaveNumE-WaveNumS)/DelWaveNum ) ) then
          iNum = int( (WaveNumE-WaveNumS)/DelWaveNum ) - ( (NProcs-1) - n )
        end if
        iNum = int( max( LParam%wnlc0(l) - WaveNumS, 0.0d0 ) / DelWaveNum )

        an_WaveNumBnds(2,n) = WaveNumS + iNum * DelWaveNum
      end if
    end do
    ! check
    do n = 0, NProcs-1
!!$      if ( MyRank == 0 ) &
!!$        write( 6, * ) n, int(an_WaveNumBnds(1,n)), int(an_WaveNumBnds(2,n))
      if ( an_WaveNumBnds(1,n) > an_WaveNumBnds(2,n) ) then
        write( 6, * ) 'Upper bound of wavenumber is less than its lower bound.'
        write( 6, * ) an_WaveNumBnds(1,n), an_WaveNumBnds(2,n)
        stop
      end if
    end do
!!$    call sleep( 10 )
!!$    stop

    WaveNumS = an_WaveNumBnds(1,MyRank)
    WaveNumE = an_WaveNumBnds(2,MyRank)

!!$    if ( WaveNumS == WaveNumE ) then
!!$      write( 6, * ) 'Unexpected zero wavenumber range for process ', MyRank
!!$      stop
!!$    end if

    call lbl_lineparam_finalize( LParam )


  end subroutine ACCalcUtilsModifyWavenumberRange2

  !----------------------------------------------------------------------------

  subroutine ACCalcUtilsModifyWavenumberRange3( &
    & FlagNetCDF, NHIT, a_HITFileName, &
    & NProcs, MyRank, &
    & NMolNum, m_MolNum, m_LineCutOffWaveNum, &
    & DelWaveNum, &
    & WaveNumS, WaveNumE &
    & )

    use lbl_module

    logical         , intent(in   ) :: FlagNetCDF
    integer         , intent(in   ) :: NHIT
    character(len=*), intent(in   ) :: a_HITFileName( NHIT )
    integer         , intent(in   ) :: NProcs
    integer         , intent(in   ) :: MyRank
    integer         , intent(in   ) :: NMolNum
    integer         , intent(in   ) :: m_MolNum(NMolNum)
    real(dp)        , intent(in   ) :: m_LineCutOffWaveNum(NMolNum)
    real(dp)        , intent(in   ) :: DelWaveNum
    real(dp)        , intent(inout) :: WaveNumS
    real(dp)        , intent(inout) :: WaveNumE


    ! Local variables
    !
    integer, parameter :: MolNumChk = 2
    logical            :: FlagClear

    type( LineParam )         :: LParam
    character(len=extstr)     :: FileName

    integer               :: NWaveNum
    real(dp), allocatable :: w_WaveNum  (:)
    integer , allocatable :: w_NCalcLine(:)
    integer               :: NCalcLineTotal
    integer               :: NCalcLineTotalTarget
    integer               :: iWaveNum
    real(dp)              :: WaveNumRangeS
    real(dp)              :: WaveNumRangeE

    integer  :: iLineS
    integer  :: iLineE
    integer  :: NLine
    real(dp) :: an_WaveNumBnds(2,0:NProcs-1)

    integer :: l
    integer :: ll
    integer :: m
    integer :: n


    FlagClear = .false.
    do m = 1, NMolNum
      if ( m_MolNum(m) == MolNumChk ) then
        FlagClear = .true.
      end if
    end do
    if ( .not. FlagClear ) then
      write( 6, * ) 'Unable to find ', MolNumChk, 'th molecule.'
      write( 6, * ) ' ', m_MolNum
      stop
    end if


    NWaveNum = int( ( WaveNumE - WaveNumS ) / DelWaveNum ) + 1
    allocate( w_WaveNum  ( NWaveNum ) )
    allocate( w_NCalcLine( NWaveNum ) )
    do l = 1, NWaveNum
      w_WaveNum(l) = WaveNumS + (l-1) * DelWaveNum
    end do


    m = MolNumChk

    FileName = a_HITFileName( m )
    write( 6, '(a)' ) '  Line File: ' // trim( FileName )

    if ( FlagNetCDF ) then
      call lbl_readfile_netcdf( FileName, 1, LParam )
      write( 6, '(a,i9)' ) '             Number of lines : ', LParam%NLine
    else
      write( 6, * ) 'ASCII format HITRAN data file is not supported, now.'
      stop
    end if


    do l = 1, LParam%NLine
      if ( WaveNumS <= LParam%wnlc0(l) ) exit
    end do
    iLineS = l
    iLineS = min( max( iLineS, 1 ), LParam%NLine )
    do l = iLineS, LParam%NLine
      if ( WaveNumE < LParam%wnlc0(l) ) exit
    end do
    iLineE = l-1
    iLineE = min( max( iLineE, iLineS ), LParam%NLine )
    NLine = iLineE - iLineS + 1


    iLineS = 1
    iLineE = iLineS

    do ll = 1, NWaveNum
      WaveNumRangeS = w_WaveNum(ll) - m_LineCutOffWaveNum(m)
      WaveNumRangeE = w_WaveNum(ll) + m_LineCutOffWaveNum(m)

      do l = iLineS, LParam%NLine
        if ( WaveNumRangeS <= LParam%wnlc0(l) ) exit
      end do
      l = min( max( l, 1 ), LParam%NLine )
      iLineS = l
      if ( LParam%wnlc0(l) <= WaveNumRangeE ) then
        do l = iLineE+1, LParam%NLine
          if ( WaveNumRangeE < LParam%wnlc0(l) ) exit
        end do
        l = l-1
        iLineE = min( l, LParam%NLine )
        if ( iLineE < iLineS ) then
          write( 6, * ) 'Unexpected iLineE which is less than iLineS.'
          stop
        end if
        iLineE = l

        w_NCalcLine(ll) = iLineE - iLineS + 1

        ! background
        w_NCalcLine(ll) = w_NCalcLine(ll) + 1000
      else
        ! There is no line in the current interval.
        w_NCalcLine(ll) = 0
      end if

    end do
    ! w_NCalcLine is cumulative number of lines used in calculation
    ! in wavenumber less than w_WaveNum
    do ll = 1+1, NWaveNum
      w_NCalcLine(ll) = w_NCalcLine(ll-1) + w_NCalcLine(ll)
    end do
    NCalcLineTotal = w_NCalcLine(NWaveNum)

!!$    if ( NLine < NProcs ) then
!!$      write( 6, * ) 'Number of processes are too large.'
!!$      write( 6, * ) '  ', NLine, NProcs
!!$      stop
!!$    end if

    iWaveNum = 1
    do n = 0, NProcs-1
      if ( n == 0 ) then
        an_WaveNumBnds(1,n) = WaveNumS
      else
        iWaveNum = iWaveNum + 1
        an_WaveNumBnds(1,n) = WaveNumS + (iWaveNum-1) * DelWaveNum
      end if
      if ( n == (NProcs-1) ) then
        an_WaveNumBnds(2,n) = WaveNumE
      else

        NCalcLineTotalTarget = int( real( NCalcLineTotal ) / NProcs * ( n+1 ) )

        do ll = iWaveNum+1, LParam%NLine
          if ( NCalcLineTotalTarget < w_NCalcLine(ll) ) exit
        end do
        iWaveNum = ll-1

        ! restrict iWaveNum so that 
        if ( ( iWaveNum + ( (NProcs-1) - n ) ) > NWaveNum ) then
          iWaveNum = NWaveNum - ( (NProcs-1) - n )
        end if

        an_WaveNumBnds(2,n) = WaveNumS + (iWaveNum-1) * DelWaveNum

      end if
    end do
    ! check
    do n = 0, NProcs-1
!!$      if ( MyRank == 0 ) &
!!$        write( 6, * ) n, int(an_WaveNumBnds(1,n)), int(an_WaveNumBnds(2,n))
      if ( an_WaveNumBnds(1,n) > an_WaveNumBnds(2,n) ) then
        write( 6, * ) 'Upper bound of wavenumber is less than its lower bound.'
        write( 6, * ) an_WaveNumBnds(1,n), an_WaveNumBnds(2,n)
        stop
      end if
    end do
!!$    call sleep( 10 )
!!$    stop

    WaveNumS = an_WaveNumBnds(1,MyRank)
    WaveNumE = an_WaveNumBnds(2,MyRank)

    call lbl_lineparam_finalize( LParam )

    deallocate( w_WaveNum   )
    deallocate( w_NCalcLine )


  end subroutine ACCalcUtilsModifyWavenumberRange3

  !----------------------------------------------------------------------------

end module ac_calc_utils
