module h2o_water

  use vtype_module
  use fi_module

  implicit none

  private

  real(dp), parameter :: RhoWater = 1.0d3

  real(dp)             , save :: a_PartSizeBndsMicron(0:3)

  integer              , save :: NSWData
  integer              , save :: NLWData
  integer              , save :: NData
  real(dp), allocatable, save :: a_WL       (:)
  real(dp), allocatable, save :: aa_ExtCoefA(:,:)
  real(dp), allocatable, save :: aa_ExtCoefB(:,:)
  real(dp), allocatable, save :: aa_ExtCoefC(:,:)
  real(dp), allocatable, save :: aa_SSCoAA  (:,:)
  real(dp), allocatable, save :: aa_SSCoAB  (:,:)
  real(dp), allocatable, save :: aa_SSCoAC  (:,:)
  real(dp), allocatable, save :: aa_AFA     (:,:)
  real(dp), allocatable, save :: aa_AFB     (:,:)
  real(dp), allocatable, save :: aa_AFC     (:,:)


  public :: H2OWaterInit
  public :: H2OWaterFinalize
  public :: H2OWaterCalcParam


  !----------------------------------------------------------------------------
contains
  !----------------------------------------------------------------------------

  subroutine H2OWaterInit( Dir )

    character(*), intent(in) :: Dir


    character(extstr) :: FileName
    character(stdstr) :: Mode = "read"

    integer           :: FU

    integer :: l

    integer           :: NHeader = 5


    a_PartSizeBndsMicron(0) =  2.5d0
    a_PartSizeBndsMicron(1) = 12.5d0
    a_PartSizeBndsMicron(2) = 30.0d0
    a_PartSizeBndsMicron(3) = 60.0d0

    FileName = trim( Dir ) // "/data_HuStamnes1993_Table1_Water_ExtCoef_SW.txt"
    NSWData = fi_lc( FileName, NHeader )
    FileName = trim( Dir ) // "/data_HuStamnes1993_Table4_Water_ExtCoef_LW.txt"
    NLWData = fi_lc( FileName, NHeader )
    NData = NSWData + NLWData
    allocate( a_WL       (NData) )
    allocate( aa_ExtCoefA(NData, 3) )
    allocate( aa_ExtCoefB(NData, 3) )
    allocate( aa_ExtCoefC(NData, 3) )
    allocate( aa_SSCoAA  (NData, 3) )
    allocate( aa_SSCoAB  (NData, 3) )
    allocate( aa_SSCoAC  (NData, 3) )
    allocate( aa_AFA     (NData, 3) )
    allocate( aa_AFB     (NData, 3) )
    allocate( aa_AFC     (NData, 3) )

    FileName = trim( Dir ) // "/data_HuStamnes1993_Table1_Water_ExtCoef_SW.txt"
    call fi_open( FileName, Mode, FU )
    do l = 1, NHeader
      read( FU, * )
    end do
    do l = 1, NSWData
      read( FU, * ) a_WL(l), &
        & aa_ExtCoefA(l,1), aa_ExtCoefB(l,1), aa_ExtCoefC(l,1), &
        & aa_ExtCoefA(l,2), aa_ExtCoefB(l,2), aa_ExtCoefC(l,2), &
        & aa_ExtCoefA(l,3), aa_ExtCoefB(l,3), aa_ExtCoefC(l,3)
    end do
    close( FU )
    FileName = trim( Dir ) // "/data_HuStamnes1993_Table4_Water_ExtCoef_LW.txt"
    call fi_open( FileName, Mode, FU )
    do l = 1, NHeader
      read( FU, * )
    end do
    do l = NSWData + 1, NData
      read( FU, * ) a_WL(l), &
        & aa_ExtCoefA(l,1), aa_ExtCoefB(l,1), aa_ExtCoefC(l,1), &
        & aa_ExtCoefA(l,2), aa_ExtCoefB(l,2), aa_ExtCoefC(l,2), &
        & aa_ExtCoefA(l,3), aa_ExtCoefB(l,3), aa_ExtCoefC(l,3)
    end do
    close( FU )

    FileName = trim( Dir ) // "/data_HuStamnes1993_Table2_Water_SSA_SW.txt"
    call fi_open( FileName, Mode, FU )
    do l = 1, NHeader
      read( FU, * )
    end do
    do l = 1, NSWData
      read( FU, * ) a_WL(l), &
        & aa_SSCoAA(l,1), aa_SSCoAB(l,1), aa_SSCoAC(l,1), &
        & aa_SSCoAA(l,2), aa_SSCoAB(l,2), aa_SSCoAC(l,2), &
        & aa_SSCoAA(l,3), aa_SSCoAB(l,3), aa_SSCoAC(l,3)
    end do
    close( FU )
    FileName = trim( Dir ) // "/data_HuStamnes1993_Table5_Water_SSA_LW.txt"
    call fi_open( FileName, Mode, FU )
    do l = 1, NHeader
      read( FU, * )
    end do
    do l = NSWData+1, NData
      read( FU, * ) a_WL(l), &
        & aa_SSCoAA(l,1), aa_SSCoAB(l,1), aa_SSCoAC(l,1), &
        & aa_SSCoAA(l,2), aa_SSCoAB(l,2), aa_SSCoAC(l,2), &
        & aa_SSCoAA(l,3), aa_SSCoAB(l,3), aa_SSCoAC(l,3)
    end do
    close( FU )

    FileName = trim( Dir ) // "/data_HuStamnes1993_Table3_Water_AF_SW.txt"
    call fi_open( FileName, Mode, FU )
    do l = 1, NHeader
      read( FU, * )
    end do
    do l = 1, NSWData
      read( FU, * ) a_WL(l), &
        & aa_AFA(l,1), aa_AFB(l,1), aa_AFC(l,1), &
        & aa_AFA(l,2), aa_AFB(l,2), aa_AFC(l,2), &
        & aa_AFA(l,3), aa_AFB(l,3), aa_AFC(l,3)
    end do
    close( FU )
    FileName = trim( Dir ) // "/data_HuStamnes1993_Table6_Water_AF_LW.txt"
    call fi_open( FileName, Mode, FU )
    do l = 1, NHeader
      read( FU, * )
    end do
    do l = NSWData+1, NData
      read( FU, * ) a_WL(l), &
        & aa_AFA(l,1), aa_AFB(l,1), aa_AFC(l,1), &
        & aa_AFA(l,2), aa_AFB(l,2), aa_AFC(l,2), &
        & aa_AFA(l,3), aa_AFB(l,3), aa_AFC(l,3)
    end do
    close( FU )



  end subroutine H2OWaterInit

  !----------------------------------------------------------------------------

  subroutine H2OWaterFinalize


    deallocate( a_WL        )
    deallocate( aa_ExtCoefA )
    deallocate( aa_ExtCoefB )
    deallocate( aa_ExtCoefC )
    deallocate( aa_SSCoAA   )
    deallocate( aa_SSCoAB   )
    deallocate( aa_SSCoAC   )
    deallocate( aa_AFA      )
    deallocate( aa_AFB      )
    deallocate( aa_AFC      )


  end subroutine H2OWaterFinalize

  !----------------------------------------------------------------------------

  subroutine H2OWaterCalcParam( &
    & PartSize, Wavelength, &
    & QExt, SSA, AF &
    & )

    real(dp), intent(in ) :: PartSize
    real(dp), intent(in ) :: Wavelength
    real(dp), intent(out) :: QExt
    real(dp), intent(out) :: SSA
    real(dp), intent(out) :: AF


    ! Local variables
    !
    real(dp) :: WavelengthMicron
    real(dp) :: PartSizeMicron
    real(dp) :: ExtCoefA
    real(dp) :: ExtCoefB
    real(dp) :: ExtCoefC
    real(dp) :: SSCoAA
    real(dp) :: SSCoAB
    real(dp) :: SSCoAC
    real(dp) :: AFA
    real(dp) :: AFB
    real(dp) :: AFC

    real(dp) :: SSCoA

    real(dp) :: a_QExt(2)
    real(dp) :: a_SSA (2)
    real(dp) :: a_AF  (2)

    integer  :: l
    integer  :: m
    integer  :: n


    PartSizeMicron   = PartSize   * 1.0d6
    WavelengthMicron = Wavelength * 1.0d6

    if ( PartSizeMicron < a_PartSizeBndsMicron(0) ) then
      PartSizeMicron   = a_PartSizeBndsMicron(0)
      n = 1
    else if ( a_PartSizeBndsMicron(3) < PartSizeMicron ) then
      PartSizeMicron   = a_PartSizeBndsMicron(3)
      n = 3
    else
      do n = 1, 3
        if ( PartSizeMicron <= a_PartSizeBndsMicron(n) ) then
          exit
        end if
      end do
    end if


    if ( WavelengthMicron <= a_WL(1) ) then
      l = 1
      !
      ExtCoefA = aa_ExtCoefA(l,n)
      ExtCoefB = aa_ExtCoefB(l,n)
      ExtCoefC = aa_ExtCoefC(l,n)
      SSCoAA   = aa_SSCoAA  (l,n)
      SSCoAB   = aa_SSCoAB  (l,n)
      SSCoAC   = aa_SSCoAC  (l,n)
      AFA      = aa_AFA     (l,n)
      AFB      = aa_AFB     (l,n)
      AFC      = aa_AFC     (l,n)
      !
      QExt  = ExtCoefA * PartSizeMicron**ExtCoefB + ExtCoefC
      SSCoA = SSCoAA   * PartSizeMicron**SSCoAB   + SSCoAC
      AF    = AFA      * PartSizeMicron**AFB      + AFC
      QExt  = QExt * 4.0d0 / 3.0d0 * RhoWater * PartSize !* 1.0d-3
      SSA   = 1.0d0 - SSCoA
    else if ( a_WL(NData) <= WavelengthMicron ) then
      l = NData
      !
      ExtCoefA = aa_ExtCoefA(l,n)
      ExtCoefB = aa_ExtCoefB(l,n)
      ExtCoefC = aa_ExtCoefC(l,n)
      SSCoAA   = aa_SSCoAA  (l,n)
      SSCoAB   = aa_SSCoAB  (l,n)
      SSCoAC   = aa_SSCoAC  (l,n)
      AFA      = aa_AFA     (l,n)
      AFB      = aa_AFB     (l,n)
      AFC      = aa_AFC     (l,n)
      !
      QExt  = ExtCoefA * PartSizeMicron**ExtCoefB + ExtCoefC
      SSCoA = SSCoAA   * PartSizeMicron**SSCoAB   + SSCoAC
      AF    = AFA      * PartSizeMicron**AFB      + AFC
      QExt  = QExt * 4.0d0 / 3.0d0 * RhoWater * PartSize !* 1.0d-3
      SSA   = 1.0d0 - SSCoA
    else

      do l = 2, NData
        if ( WavelengthMicron <= a_WL(l) ) exit
      end do
      !
      do m = 1, 2
        ExtCoefA = aa_ExtCoefA(l-2+m,n)
        ExtCoefB = aa_ExtCoefB(l-2+m,n)
        ExtCoefC = aa_ExtCoefC(l-2+m,n)
        SSCoAA   = aa_SSCoAA  (l-2+m,n)
        SSCoAB   = aa_SSCoAB  (l-2+m,n)
        SSCoAC   = aa_SSCoAC  (l-2+m,n)
        AFA      = aa_AFA     (l-2+m,n)
        AFB      = aa_AFB     (l-2+m,n)
        AFC      = aa_AFC     (l-2+m,n)
        !
        QExt  = ExtCoefA * PartSizeMicron**ExtCoefB + ExtCoefC
        SSCoA = SSCoAA   * PartSizeMicron**SSCoAB   + SSCoAC
        AF    = AFA      * PartSizeMicron**AFB      + AFC
        QExt  = QExt * 4.0d0 / 3.0d0 * RhoWater * PartSize !* 1.0d-3
        SSA   = 1.0d0 - SSCoA
        !
        a_QExt(m) = QExt
        a_SSA (m) = SSA
        a_AF  (m) = AF
      end do

      QExt = LinearInterpolate( a_WL(l-1:l), a_QExt, WavelengthMicron )
      SSA  = LinearInterpolate( a_WL(l-1:l), a_SSA , WavelengthMicron )
      AF   = LinearInterpolate( a_WL(l-1:l), a_AF  , WavelengthMicron )

    end if

    QExt  = max( QExt, 0.0d0 )
    SSA   = max( SSA , 0.0d0 )
    AF    = max( AF  , 0.0d0 )

  end subroutine H2OWaterCalcParam

  !----------------------------------------------------------------------------

  function LinearInterpolate( a_X, a_Y, X ) result( Y )

    real(dp), intent(in ) :: a_X(2)
    real(dp), intent(in ) :: a_Y(2)
    real(dp), intent(in ) :: X

    real(dp) :: Y


    Y = ( a_Y(2) - a_Y(1) ) / ( a_X(2) - a_X(1) ) * ( X - a_X(1) ) + a_Y(1)

  end function LinearInterpolate

  !----------------------------------------------------------------------------

end module h2o_water
