!= 
!
!= Wrapper program for twostr
!
module rad_rte_twostr_wrapper_v1
  !
  != 
  !
  != Wrapper program for twostr
  !
  ! <b>Note that Japanese and English are described in parallel.</b>
  !
  ! 
  !
  !
  !== References
  !
!!$  !  Toon, O. B., C. P. McKay, and A. P. Ackerman, 
!!$  !    Rapid calculation of radiative heating rates and photodissociation rates 
!!$  !    in inhomogeneous multiple scattering atmospheres, 
!!$  !    J. Geophys. Res., 94, 16287-16301, 1989.
  !
  !== Procedures List
  !
!!$  ! RadiationFluxDennouAGCM :: 放射フラックスの計算
!!$  ! RadiationDTempDt        :: 放射フラックスによる温度変化の計算
!!$  ! RadiationFluxOutput     :: 放射フラックスの出力
!!$  ! RadiationFinalize       :: 終了処理 (モジュール内部の変数の割り付け解除)
!!$  ! ------------            :: ------------
!!$  ! RadiationFluxDennouAGCM :: Calculate radiation flux
!!$  ! RadiationDTempDt        :: Calculate temperature tendency with radiation flux
!!$  ! RadiationFluxOutput     :: Output radiation fluxes
!!$  ! RadiationFinalize       :: Termination (deallocate variables in this module)
  !
  !== NAMELIST
  !
!!$  ! NAMELIST#radiation_DennouAGCM_nml
  !

  use vtype_module

  ! 物理・数学定数設定
  ! Physical and mathematical constants settings
  !
  use constants0, only: &
    & PI
                            ! $ \pi $ .
                            ! 円周率.  Circular constant


  implicit none

  private


  ! Local variables
  !


  ! 公開変数
  ! Public variables
  !
!!$  logical, save :: rad_rte_twostr_wrapper_v1_inited = .false.
  logical, save :: rad_rte_twostr_wrapper_v1_inited = .true.
                              ! 初期設定フラグ.
                              ! Initialization flag

  public :: RadRTETWOSTRWrapper1DSW
  public :: RadRTETWOSTRWrapper1DLW
  public :: RadRTETWOSTRWrapper1D


  ! INTERFACE 文 ; INTERFACE statements
  !
!!$  interface RadRTETwoStreamApp
!!$    module procedure RadRTETwoStreamAppWrapper, RadRTETwoStreamAppCore
!!$  end interface
!!$  interface RadRTETwoStreamApp
!!$    module procedure RadRTETwoStreamAppWrapper
!!$  end interface



  character(*), parameter:: module_name = 'rad_rte_twostr_wrapper_v1'
                              ! モジュールの名称.
                              ! Module name
  character(*), parameter:: version = &
    & '$Name:  $' // &
    & '$Id: rad_rte_two_stream_app.f90,v 1.7 2015/03/11 04:48:47 yot Exp $'
                              ! モジュールのバージョン
                              ! Module version

  !--------------------------------------------------------------------------------------

contains

  !--------------------------------------------------------------------------------------

  subroutine RadRTETWOSTRWrapper1DSW(    &
    & kmax,                         & ! (in)
    & z_SSA, z_AF,                  & ! (in)
    & r_OptDep,                     & ! (in)
    & SurfAlbedo,                   & ! (in)
    & SolarFluxTOA, CosSZA,         & ! (in)
    & r_RadUwFlux, r_RadDwFlux      & ! (out)
    & )

    ! USE statements
    !

    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: z_SSA      (1:kmax)
    real(DP), intent(in ) :: z_AF       (1:kmax)
    real(DP), intent(in ) :: r_OptDep   (0:kmax)
    real(DP), intent(in ) :: SurfAlbedo
    real(DP), intent(in ) :: SolarFluxTOA
    real(DP), intent(in ) :: CosSZA
                              ! sec (入射角).
                              ! sec (angle of incidence)
    real(DP), intent(out) :: r_RadUwFlux(0:kmax)
    real(DP), intent(out) :: r_RadDwFlux(0:kmax)

    ! Local variables
    !
    real(DP) :: WaveNum
    logical  :: FlagEmis
    real(dp) :: r_Temp(0:kmax)
    real(dp) :: SurfTemp
    integer , parameter :: NDelFlxMax = 1
    real(DP)            :: ra_DelRadUwFlux(0:kmax, -1:NDelFlxMax)
    real(DP)            :: ra_DelRadDwFlux(0:kmax, -1:NDelFlxMax)


    ! 初期化確認
    ! Initialization check
    !
    if ( .not. rad_rte_twostr_wrapper_v1_inited ) then
      write( 6, * ) "Error: ", trim( module_name ), &
        & 'This module has not been initialized.'
      stop
    end if


    if ( CosSZA > 0.0d0 ) then
      ! Daytime

      FlagEmis = .false.
      WaveNum  = 0.0d0
      r_Temp   = 0.0d0
      SurfTemp = 0.0d0

      call RadRTETWOSTRWrapper1D(                          &
        & FlagEmis,                                              & ! (in)
        & WaveNum, WaveNum,                                      & ! (in)
        & kmax,                                                  & ! (in)
        & z_SSA, z_AF,                                           & ! (in)
        & r_OptDep,                                              & ! (in)
        & SurfAlbedo,                                            & ! (in)
        & SolarFluxTOA, CosSZA,                                  & ! (in)
        & r_Temp, SurfTemp,                                      & ! (in)
        & NDelFlxMax,                                            & ! (in)
        & r_RadUwFlux, r_RadDwFlux,                              & ! (out)
        & ra_DelRadUwFlux, ra_DelRadDwFlux                       & ! (out)
        & )

    else
      r_RadUwFlux = 0.0d0
      r_RadDwFlux = 0.0d0
    end if


  end subroutine RadRTETWOSTRWrapper1DSW

  !------------------------------------------------------------------------------------

  subroutine RadRTETWOSTRWrapper1DLW(          &
    & WaveNumS, WaveNumE,                      & ! (in)
    & kmax,                                    & ! (in)
    & z_SSA, z_AF,                             & ! (in)
    & r_OptDep,                                & ! (in)
    & SurfAlbedo,                              & ! (in)
    & r_Temp, SurfTemp,                        & ! (in)
    & NDelFlxMax,                              & ! (in)
    & r_RadUwFlux, r_RadDwFlux,                & ! (out)
    & ra_DelRadUwFlux, ra_DelRadDwFlux         & ! (out)
    & )

    ! USE statements
    !

    real(DP), intent(in ) :: WaveNumS
    real(DP), intent(in ) :: WaveNumE
    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: z_SSA          (1:kmax)
    real(DP), intent(in ) :: z_AF           (1:kmax)
    real(DP), intent(in ) :: r_OptDep       (0:kmax)
    real(DP), intent(in ) :: SurfAlbedo
    real(DP), intent(in ) :: r_Temp         (0:kmax)
    real(DP), intent(in ) :: SurfTemp
    integer , intent(in ) :: NDelFlxMax
    real(DP), intent(out) :: r_RadUwFlux    (0:kmax)
    real(DP), intent(out) :: r_RadDwFlux    (0:kmax)
    real(DP), intent(out) :: ra_DelRadUwFlux(0:kmax, -1:NDelFlxMax)
    real(DP), intent(out) :: ra_DelRadDwFlux(0:kmax, -1:NDelFlxMax)


    ! Local variables
    !
    logical  :: FlagEmis
    real(DP) :: SolarFluxTOA
    real(DP) :: CosSZA


    ! 初期化確認
    ! Initialization check
    !
    if ( .not. rad_rte_twostr_wrapper_v1_inited ) then
      write( 6, * ) "Error: ", trim( module_name ), &
        & 'This module has not been initialized.'
      stop
    end if


    FlagEmis     = .true.
    SolarFluxTOA = 0.0d0
    CosSZA       = 0.0d0

    call RadRTETWOSTRWrapper1D(                                &
      & FlagEmis,                                              & ! (in)
      & WaveNumS, WaveNumE,                                    & ! (in)
      & kmax,                                                  & ! (in)
      & z_SSA, z_AF,                                           & ! (in)
      & r_OptDep,                                              & ! (in)
      & SurfAlbedo,                                            & ! (in)
      & SolarFluxTOA, CosSZA,                                  & ! (in)
      & r_Temp, SurfTemp,                                      & ! (in)
      & NDelFlxMax,                                            & ! (in)
      & r_RadUwFlux, r_RadDwFlux,                              & ! (out)
      & ra_DelRadUwFlux, ra_DelRadDwFlux                       & ! (out)
      & )


  end subroutine RadRTETWOSTRWrapper1DLW

  !--------------------------------------------------------------------------------------

  subroutine RadRTETWOSTRWrapper1D(                          &
    & FlagEmis,                                              & ! (in)
    & WaveNumS, WaveNumE,                                    & ! (in)
    & kmax,                                                  & ! (in)
    & z_SSA, z_AF,                                           & ! (in)
    & r_OptDep,                                              & ! (in)
    & SurfAlbedo,                                            & ! (in)
    & SolarFluxTOA, CosSZA,                                  & ! (in)
    & r_Temp, SurfTemp,                                      & ! (in)
    & NDelFlxMax,                                            & ! (in)
    & r_RadUwFlux, r_RadDwFlux,                              & ! (out)
    & ra_DelRadUwFlux, ra_DelRadDwFlux                       & ! (out)
    & )

    ! USE statements
    !


    logical , intent(in ) :: FlagEmis
    real(DP), intent(in ) :: WaveNumS
    real(DP), intent(in ) :: WaveNumE
    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: z_SSA      (1:kmax)
    real(DP), intent(in ) :: z_AF       (1:kmax)
    real(DP), intent(in ) :: r_OptDep   (0:kmax)
    real(DP), intent(in ) :: SurfAlbedo
    real(DP), intent(in ) :: SolarFluxTOA
    real(DP), intent(in ) :: CosSZA
                              ! sec (入射角).
                              ! sec (angle of incidence)
    real(DP), intent(in ) :: r_Temp         (0:kmax)
    real(DP), intent(in ) :: SurfTemp
    integer , intent(in ) :: NDelFlxMax
    real(DP), intent(out) :: r_RadUwFlux(0:kmax)
    real(DP), intent(out) :: r_RadDwFlux(0:kmax)
    real(DP), intent(out) :: ra_DelRadUwFlux(0:kmax, -1:NDelFlxMax)
    real(DP), intent(out) :: ra_DelRadDwFlux(0:kmax, -1:NDelFlxMax)


    ! Local variables
    !

    !
    ! SSAAdj    : Delta-Function Adjusted Single-Scattering Albedo
    ! AFAdj     : Delta-Function Adjusted Asymmetry Factor
    ! OpDepAdj  : Delta-Function Adjusted Optical Depth
    !
    real(DP) :: z_SSAAdj     ( 1:kmax )
    real(DP) :: z_AFAdj      ( 1:kmax )
    real(DP) :: r_OpDepAdj   ( 0:kmax )


    !
    ! temporary variables
    !
    real(dp) :: z_DelOptDep          (1:kmax)

    real(sp) :: z_DelOptDepInvOrd4   (1:kmax)
    real(sp) :: z_AFInvOrd4          (1:kmax)
    real(sp) :: z_SSAInvOrd4         (1:kmax)
    real(sp) :: r_TempInvOrd4        (0:kmax)
    real(sp) :: r_OptDepInvOrd4      (0:kmax)
    real(sp) :: r_ZInvOrd4           (0:kmax)
    real(sp) :: r_DFDOptDepInvOrd4   (0:kmax)
    real(sp) :: r_RadUwFluxInvOrd4   (0:kmax)
    real(sp) :: r_RadDwFluxDirInvOrd4(0:kmax)
    real(sp) :: r_RadDwFluxDifInvOrd4(0:kmax)
    real(sp) :: r_RadDwFluxInvOrd4   (0:kmax)
    real(sp) :: r_RadMeanIntInvOrd4  (0:kmax)


    real(sp) :: SurfAlbedo4
    real(sp) :: SurfTemp4
    real(sp) :: SolarFluxTOA4
    real(sp) :: CosSZA4
    real(sp) :: WaveNumS4
    real(sp) :: WaveNumE4

    character(127)    :: Header = "###"
    integer           :: a_iError(22)
    logical           :: FlagDeltaM = .true.
!!$    logical           :: FlagDeltaM = .false.
    logical           :: a_FlagPrnt(2) = .false.
!!$    logical           :: a_FlagPrnt(2) = .true.
    logical           :: FlagQuiet  = .true.
!!$    logical           :: FlagQuiet  = .false.


    integer  :: k


    ! 初期化確認
    ! Initialization check
    !
    if ( .not. rad_rte_twostr_wrapper_v1_inited ) then
      write( 6, * ) "Error: ", trim( module_name ), &
        & 'This module has not been initialized.'
      stop
    end if


    !
    ! Delta-Function Adjustment
    !
    if ( .false. ) then
      z_AFAdj  = z_AF / ( 1.0d0 + z_AF )
      z_SSAAdj =   ( 1.0d0 - z_AF**2 ) * z_SSA / ( 1.0d0 - z_SSA * z_AF**2 )
      !
      r_OpDepAdj = 0.0d0
      do k = kmax-1, 0, -1
        r_OpDepAdj(k) =                                 &
          &   r_OpDepAdj(k+1)                           &
          & + ( 1.0d0 - z_SSA(k+1) * z_AF(k+1)**2 )     &
          &   * ( r_OptDep(k) - r_OptDep(k+1) )
      end do
    else
      z_AFAdj  = z_AF
      z_SSAAdj = z_SSA
      !
      r_OpDepAdj = r_OptDep
    end if


    do k = 1, kmax
      z_DelOptDep(k) = r_OpDepAdj(k-1) - r_OpDepAdj(k)
    end do


    do k = 1, kmax
      z_DelOptDepInvOrd4(k) = real( z_DelOptDep(kmax-k+1) )
      z_AFInvOrd4       (k) = real( z_AFAdj    (kmax-k+1) )
      z_SSAInvOrd4      (k) = real( z_SSAAdj   (kmax-k+1) )
    end do
    do k = 0, kmax
      r_TempInvOrd4  (k) = real( r_Temp  (kmax-k) )
      r_OptDepInvOrd4(k) = real( r_OptDep(kmax-k) )
    end do

    r_ZInvOrd4 = 0.0d0


    SurfAlbedo4   = real( SurfAlbedo   ) * ( sqrt( 3.0 ) / 2.0 )
    SurfTemp4     = real( SurfTemp     )
    SolarFluxTOA4 = real( SolarFluxTOA )
    CosSZA4       = real( CosSZA       )
    WaveNumS4     = real( WaveNumS * 1.0d-2 )
    WaveNumE4     = real( WaveNumE * 1.0d-2 )
!!$    SurfAlbedo4   = SurfAlbedo4 * 0.2 / 0.2309400
    call twostr( &
      & SurfAlbedo4, & ! albedo
      & SurfTemp4, & ! btemp
      & FlagDeltaM, & ! deltam
      & z_DelOptDepInvOrd4, & ! dtauc
      & SolarFluxTOA4, & ! fbeam
      & 0.0, & ! fisot
      & z_AFInvOrd4, & ! gg
      & Header, & ! header
      & a_iError, & ! ierror
      & kmax, & ! maxcly
      & kmax+1, & ! maxulv
      & kmax, & ! nlyr
      & FlagEmis, & ! planck
      & kmax+1, & ! ntau
      & a_FlagPrnt, & ! prnt
      & FlagQuiet, & ! quiet
      & 0.0, & ! radius
      & .false., & ! spher
      & z_SSAInvOrd4, & ! ssalb
      & 0.0, & ! temis
      & r_TempInvOrd4, & ! temper
      & r_TempInvOrd4(0), & ! ttemp
      & CosSZA4, & ! umu0
      & .false., & ! usrtau
      & r_OptDepInvOrd4, & ! utau
      & WaveNumS4, & ! wvnmlo
      & WaveNumE4, & ! wvnmhi
      & r_ZInvOrd4, & ! zd
      & r_DFDOptDepInvOrd4, & ! dfdt
      & r_RadUwFluxInvOrd4, & ! flup
      & r_RadDwFluxDirInvOrd4, & ! rfldir
      & r_RadDwFluxDifInvOrd4, & ! rfldn
      & r_RadMeanIntInvOrd4 & ! uavg
      & )
    if ( any( a_iError /= 0 ) ) then
      write( 6, * ) 'TWOSTR found errors: ', a_iError
      stop
    end if

    r_RadDwFluxInvOrd4 = r_RadDwFluxDirInvOrd4 + r_RadDwFluxDifInvOrd4

    do k = 0, kmax
      r_RadUwFlux(k) = dble( r_RadUwFluxInvOrd4(kmax-k) )
      r_RadDwFlux(k) = dble( r_RadDwFluxInvOrd4(kmax-k) )
    end do


    ra_DelRadUwFlux = 0.0d0
    ra_DelRadDwFlux = 0.0d0


  end subroutine RadRTETWOSTRWrapper1D

  !----------------------------------------------------------------------------

!!$  subroutine RadRTETwoStreamAppInit

!!$    ! ファイル入出力補助
!!$    ! File I/O support
!!$    !
!!$    use dc_iounit, only: FileOpen
!!$
!!$    ! NAMELIST ファイル入力に関するユーティリティ
!!$    ! Utilities for NAMELIST file input
!!$    !
!!$    use namelist_util, only: namelist_filename, NmlutilMsg, NmlutilAryValid



    ! 宣言文 ; Declaration statements
    !

!!$    integer:: unit_nml        ! NAMELIST ファイルオープン用装置番号.
!!$                              ! Unit number for NAMELIST file open
!!$    integer:: iostat_nml      ! NAMELIST 読み込み時の IOSTAT.
!!$                              ! IOSTAT of NAMELIST read

    ! NAMELIST 変数群
    ! NAMELIST group name
    !
!!$    namelist /rad_rte_two_stream_app_v2_nml/ &
!!$      & NGaussQuad
          !
          ! デフォルト値については初期化手続 "rad_rte_two_stream_app#RadRTETwoStreamAppInit"
          ! のソースコードを参照のこと.
          !
          ! Refer to source codes in the initialization procedure
          ! "rad_rte_two_stream_app#RadRTETwoStreamAppInit" for the default values.
          !

!!$    if ( rad_rte_two_stream_app_v3_inited ) return


    ! デフォルト値の設定
    ! Default values settings
    !
!!$    NGaussQuad = 8


    ! NAMELIST の読み込み
    ! NAMELIST is input
    !
!!$    if ( trim(namelist_filename) /= '' ) then
!!$      call FileOpen( unit_nml, &          ! (out)
!!$        & namelist_filename, mode = 'r' ) ! (in)
!!$
!!$      rewind( unit_nml )
!!$      read( unit_nml,                          & ! (in)
!!$        & nml = rad_rte_two_stream_app_nml,    & ! (out)
!!$        & iostat = iostat_nml )                  ! (out)
!!$      close( unit_nml )
!!$
!!$      call NmlutilMsg( iostat_nml, module_name ) ! (in)
!!$    end if


!!$    allocate( a_GQP(1:NGaussQuad) )
!!$    allocate( a_GQW(1:NGaussQuad) )


    ! Initialization of modules used in this module
    !


    ! 印字 ; Print
    !
!!$    call MessageNotify( 'M', module_name, '----- Initialization Messages -----' )
!!$    call MessageNotify( 'M', module_name, 'NGaussQuad = %d', i = (/ NGaussQuad /) )
!!$    call MessageNotify( 'M', module_name, '-- version = %c', c1 = trim(version) )


!!$    rad_rte_two_stream_app_v3_inited = .true.
!!$
!!$
!!$  end subroutine RadRTETwoStreamAppInit

  !----------------------------------------------------------------------------

end module rad_rte_twostr_wrapper_v1
