module saturation_H2O_T1930

  use vtype_module

  implicit none

  public :: CalcPressH2OVapSat
  public :: CalcDPressH2OVapSatDTemp

contains

  !----------------------------------------------------------------------------

  function CalcPressH2OVapSat( Temp ) result( PressH2OVapSat )

    use constants0, only : GasRUniv

    real(DP), intent(in ) :: Temp

    real(DP) :: PressH2OVapSat


    ! NHA1992
!!$    PressH2OVapSat = 1.4d11 * exp( - LatentHeatPerMol / ( GasRUniv * Temp ) )

    ! Tetens
    if ( Temp < 273.16d0 ) then
      PressH2OVapSat = &
        & 6.1078e2 * exp( 21.8745584d0 * ( Temp - 273.16d0 ) / ( Temp -  7.66d0 ) )
    else
      PressH2OVapSat = &
        & 6.1078e2 * exp( 17.2693882d0 * ( Temp - 273.16d0 ) / ( Temp - 35.86d0 ) )
    end if


  end function CalcPressH2OVapSat

  !----------------------------------------------------------------------------

  function CalcDPressH2OVapSatDTemp( Temp ) result( DPressH2OVapSatDTemp )

    use constants0, only : GasRUniv

    real(DP), intent(in ) :: Temp

    real(DP) :: DPressH2OVapSatDTemp


    ! NHA1992
!!$    DPressH2OVapSatDTemp = &
!!$      & 1.4d11 * exp( - LatentHeatPerMol / ( GasRUniv * Temp ) ) &
!!$      & * LatentHeatPerMol / ( GasRUniv * Temp**2 )

    ! Tetens
    if ( Temp < 273.16d0 ) then
      DPressH2OVapSatDTemp = &
        & 6.1078e2 * exp( 21.8745584d0 * ( Temp - 273.16d0 ) / ( Temp -  7.66d0 ) ) &
        & * 21.8745584d0 &
        & * ( ( Temp -  7.66d0 ) - ( Temp - 273.16d0 ) ) / ( Temp -  7.66d0 )**2
    else
      DPressH2OVapSatDTemp = &
        & 6.1078e2 * exp( 17.2693882d0 * ( Temp - 273.16d0 ) / ( Temp - 35.86d0 ) ) &
        & * 17.2693882d0 &
        & * ( ( Temp - 35.86d0 ) - ( Temp - 273.16d0 ) ) / ( Temp - 35.86d0 )**2
    end if


  end function CalcDPressH2OVapSatDTemp

  !----------------------------------------------------------------------------

end module saturation_H2O_T1930
